source mkcolor2.tcl
proc tk_bindcpy { destclass srclass } {
  foreach i [bind $srclass] {
    bind $destclass $i [bind $srclass $i]
  }
}
set tconfig_bg #eeeeee
option add *TConfigButton.background "$tconfig_bg"
option add *TConfigRadiobutton.background "$tconfig_bg"
option add *TConfigLabel.background "$tconfig_bg"
option add *TConfigDialog.background "$tconfig_bg"
option add *TConfigFrame.background "$tconfig_bg"
option add *TConfigListbox.background "$tconfig_bg"
option add *TConfigEntry.background "#ffffff"
option add *TConfigEntry.font "System-Bold-14"
tk_bindcpy TConfigButton Button
tk_bindcpy TConfigRadiobutton Radiobutton
tk_bindcpy TConfigListbox Listbox
tk_bindcpy TConfigEntry Entry



proc check_keys { keyfile doproc keytab} {
    if { [ catch { source $keyfile } msg ] == 1 } {
      msgbox . [format "%s,during reading $keyfile" $msg] Error
      return
    }
    if { [ catch { eval "$doproc $keytab" } msg ] == 1 } {
      msgbox . [format "%s,during execute $keyfile" $msg] Error
    }
}
proc switch_keys { keytype } {
  puts stdout "switch_keys $keytype"
  switch $keytype {
    vt100 { check_keys keyvt.tcl dovtkeys wtkeytab }
    ansi  { check_keys keyansi.tcl doansikeys wtkeytab }
    xterm { check_keys keyxterm.tcl doxtermkeys wtkeytab }
    user  { check_keys keyuser.tcl douserkeys wtkeytab }
  }
}


proc tsaveconfig { term keytab file } {
  global $keytab
  if { [info commands $term] == "" } {
    msgbox . "$term is no terminal in saveconfig" Error
    return
  }
  if { [catch {set fileid [open $file w]} msg] } {
    msgbox . $msg Error
    return
  }
  puts $fileid "proc user_config_term { term keytab } {"
#save screen position
  scan [wm geometry .] "%dx%d+%d+%d" cx cy x y
#  puts $fileid "wm geometry . ${cx}x$cy+$x+$y"
  puts $fileid "wm geometry . +$x+$y"
#save configure options
  puts $fileid "#  configure options"
  foreach i [lsort [$term configure]] {
    if { [lindex $i 4] != "" } {
      puts $fileid "  \$term configure [lindex $i 0] \"[lindex $i 4]\""
    }
  }
#save commands
  puts $fileid "#  commands"
  foreach i [lsort [$term attsattsall]] {
    puts $fileid "  \$term attsatts \"$i\" \"[$term attsatts $i]\""
  }
  puts $fileid "  \$term wrap [$term wrap]"
  puts $fileid "#  keytable for the terminal"
  puts $fileid "  global \$keytab"
  foreach i [lsort [array names $keytab]] {
    puts $fileid "  set \$\{keytab\}\($i\) \"[interpretchars [set ${keytab}($i)]]\""
  }
  puts $fileid "  switch_keys \[set \$\{keytab\}(TYPE)\]"
  puts $fileid "  return ok"
  puts $fileid "}"
  close $fileid
}
#===========================================================
# menut :
#	Defines a the menus for the terminal configuration
# configuration
# Parameter:	root : is the frame witch is the father of
#			terminal.
#		term : is the full path of the frame terminal
#-----------------------------------------------------------
proc menut { root term background } {
  frame $root.m -class menuframe 
  place $root.m -x 0
  menubutton $root.m.config -menu $root.m.config.m -text "Configuration"
  place $root.m.config -x 0
  menu $root.m.config.m
  $root.m.config.m add command -label "Colors..." \
     -command "showColors $root $term $background"
  $root.m.config.m add command -label "Fonts..." \
     -command "chooseFonts $root $term $background"
  $root.m.config.m add command -label "Attributes..." \
     -command "showAtts $root $term $background"
  $root.m.config.m add command -label "Terminal-specific..." \
     -command "showSpecs $root $term $background"
  $root.m.config.m add command -label "Save Configuration" \
     -command "tsaveconfig $term wtkeytab termuser.tcl"
  $root.m.config.m add command -label "Reset" \
     -command "newparse"
  $root.m.config.m add command -label "Dump mem.log" \
     -command "memory active mem.log"
  $root.m.config.m add command -label "current allocated" \
     -command {msgbox . "current allocated:[memory current_malloced]" "Memory Info"}
}

proc chooseFonts { root term background} {
  if { [set font [choosefont $root {fixedpitchonly}]] != "" } {
    $term configure -font $font
  }
}

#===========================================================
# showColors :
# Parameter: root : is the frame witch is the father of
#		terminal.
#-----------------------------------------------------------

proc showColors { root term background} {
  if {! [winfo exists $root.l_colors]} {
     toplevel $root.l_colors -class TConfigDialog -relief groove -relief raised -bd 1
     listbox $root.l_colors.lb -class TConfigListbox -relief sunken -bd 1
          $root.l_colors.lb insert end "yellow"
          $root.l_colors.lb insert end "red"
          $root.l_colors.lb insert end "black"
          $root.l_colors.lb insert end "white"
          $root.l_colors.lb insert end "blue"
          $root.l_colors.lb insert end "magenta"
          $root.l_colors.lb insert end "cyan"
          $root.l_colors.lb insert end "green"
     bind $root.l_colors.lb <Double-Button-1> "showThisColor $root $term $background"
     button $root.l_colors.ok -class TConfigButton -text "Ok" \
           -command "showThisColor $root $term $background"

     button $root.l_colors.nil -class TConfigButton -text "Abbrechen" \
           -command "destroy $root.l_colors;focus $root"

     pack $root.l_colors.lb -side top -pady 2 -padx 2 -fill y -expand 1
     pack $root.l_colors.ok $root.l_colors.nil -side left -pady 2 -padx 2 \
                  -fill x -expand 1
     wm title $root.l_colors "Colors"
  } else {
     raise $root.l_colors $root
  }
}
#===========================================================
# Subroutines for Attribut-editing
#===========================================================


#callback if pressing on a color radiobutton
proc attsattscolor { term color fg } {
  global actatts
  if {$fg=="fg"} {
  } else {
     $term attsatts $actatts $color
  }
}

#callback if pressing on a relief radiobutton
proc attsattsrelief { term relief } {
  global actatts
  $term attsatts $actatts $relief
}

#creation of a column title
proc createTitle { parent name relief bd} {
  frame $parent.f -class TConfigFrame -relief $relief -bd $bd
  label $parent.f.title -class TConfigLabel -text $name
  pack $parent.f -fill x -side top
  pack $parent.f.title -fill x -side top
}


proc cloneTerminal { newterm term } {
  set termstr "term \{$newterm\}"
  #copy all options
  foreach i [lsort [$term configure]] {
    if { [lindex $i 4] != "" } {
      set str " [lindex $i 0] \"[lindex $i 4]\""
      append termstr $str
    }
  }
  #create the terminal
  eval $termstr
  #copy attributes
  foreach i [$term attsattsall] {
    $newterm attsatts $i [$term attsatts $i]
  }
  #  $newterm configure -relief [lindex [$term configure -relief] 4]
}
proc attsClick { cterm xpos ypos } {
  set pos [$cterm nearest $xpos $ypos]
  scan $pos "%d %d" col line
  $cterm goto 0 $line
  $cterm cursor on
  global actatts
  set actatts [$cterm cursoratts]
  set attsatts [$cterm attsatts $actatts]
  scan $attsatts "%s %s" relief bgcolor
  global actattsrelief
  set actattsrelief $relief
  global actattsbgcolor
  set actattsbgcolor $bgcolor
}
proc insertAttsLine { parent term atts name } {
  set cterm $parent.t
  if { [info commands $cterm]== ""  } {
    cloneTerminal $cterm $term
    $cterm configure -height 0
    $cterm configure -width 10
    $cterm cursor off
    bind $cterm <1> { attsClick %W %x %y }
    pack $cterm
  }
  set lines [lindex [$cterm configure -height] 4]
  incr lines
  $cterm configure -height $lines
  set cols [lindex [$cterm configure -width] 4]
  set len [string length $name]
  if {$len > $cols} {
    $cterm configure -width [expr $len+1]
  }
  $cterm goto 0 [expr $lines-1]
  $cterm attr $atts
  $cterm puts $name
  $cterm attr normal
  $cterm goto 0 [expr $lines-1]
  return $cterm
}
#creation of a radiobutton and a label inside a frame
proc createAttsItem { parent term atts name } {
  set attsatts [$term attsatts $atts]
  scan $attsatts "%s %s" relief bg
  return [insertAttsLine $parent $term $atts $name]
}

#creation of a radiobutton and a color frame inside a frame
#param fg is "fg" or "bg"
proc createColorItem { parent term color name fg} {
  frame $parent.$color -class TConfigFrame
  radiobutton $parent.$color.r -class TConfigRadiobutton \
    -width 8 \
    -command "attsattscolor $term $color $fg" \
    -variable actattsbgcolor  -value $color -text $name \
    -anchor w
#this is a special colored frame 
  frame $parent.$color.f -bg [lindex [$term configure -$color] 4]\
    -width 30 -height 20
  pack $parent.$color.r -side left -fill x
  pack $parent.$color.f -side right
  pack $parent.$color -anchor w
  return $parent.$color
}
#creation of a radiobutton and a relief frame inside a frame
proc createReliefItem { parent term relief name} {
  frame $parent.$relief -class TConfigFrame
  radiobutton $parent.$relief.r -class TConfigRadiobutton \
    -width 8 \
    -command "attsattsrelief $term $relief" \
    -variable actattsrelief  -value $relief -text $name \
    -anchor w
  frame $parent.$relief.f -class TConfigFrame\
    -relief $relief -bd 1 -width 30 -height 20
  pack $parent.$relief.r -side left -fill x
  pack $parent.$relief.f -side right
  pack $parent.$relief -anchor w
  return $parent.$relief
}
proc changeConfigAtts { term cterm } {
   foreach i [$cterm attsattsall] {
     $term attsatts $i [$cterm attsatts $i]
   }
}
proc cancelAtts { w term cterm } {
#delete all global variables
  catch { global actatts }
  catch { unset  actatts }
  global actattsbgcolor
  unset  actattsbgcolor
  global actattsrelief
  unset  actattsrelief
  catch { global attslabeltoset }
  catch { unset  attslabeltoset }
  destroy $w
  focus $term
}
proc doneAtts { w term cterm } {
  changeConfigAtts $term $cterm
  cancelAtts $w $term $cterm
}
proc doAtts { w term cterm } {
  changeConfigAtts $term $cterm
}
#===========================================================
# showAtts : Displays a Configuration-Dialog for Terminal
# Attributes like bold,underline and s.o.
#-----------------------------------------------------------

proc showAtts { root term background} {
  set w $root.l_atts
  if {! [winfo exists $w]} {
    toplevel $w -class TConfigDialog -relief groove -relief raised -bd 1
    set titlerelief flat
    set titlebd 1
    #the attributes
    frame $w.fa -class TConfigFrame -bd $titlebd -relief raised
    createTitle    $w.fa "Attribute" $titlerelief $titlebd
    createAttsItem $w.fa $term "bold" "fett"
    createAttsItem $w.fa $term "bold reverse" "fett invers"
    createAttsItem $w.fa $term "bold underline" "fett unterstrichen"
    createAttsItem $w.fa $term "bold blink"  "fett blinkend"
    createAttsItem $w.fa $term "bold reverse underline" "fett invers unterstrichen"
    createAttsItem $w.fa $term "bold reverse blink" "fett invers blinkend"
    createAttsItem $w.fa $term "bold underline blink" "fett unterstrichen blinkend"
    createAttsItem $w.fa $term "bold reverse underline blink" "fett invers unterstrichen blinkend"
    createAttsItem $w.fa $term "reverse" "invers"
    createAttsItem $w.fa $term "reverse underline" "invers unterstrichen"
    createAttsItem $w.fa $term "reverse blink"  "invers blinkend"
    createAttsItem $w.fa $term "reverse underline blink"  "invers unterstrichen blinkend"
    createAttsItem $w.fa $term "underline"  "unterstrichen"
    createAttsItem $w.fa $term "underline blink"  "unterstrichen blinkend"
    set cterm [createAttsItem $w.fa $term "blink"  "blinkend"]
    #the backgound colors
    frame $w.fc -class TConfigFrame -bd $titlebd -relief raised
    createTitle     $w.fc "Hintergrund-Farbe" $titlerelief $titlebd
    createColorItem $w.fc $cterm white weiss bg
    createColorItem $w.fc $cterm yellow gelb bg
    createColorItem $w.fc $cterm red rot bg
    createColorItem $w.fc $cterm green gruen bg
    createColorItem $w.fc $cterm blue blau bg
    createColorItem $w.fc $cterm cyan cyan bg
    createColorItem $w.fc $cterm magenta magenta bg
    createColorItem $w.fc $cterm black schwarz bg
    #the reliefs
    frame $w.fr -class TConfigFrame -bd $titlebd -relief raised
    createTitle      $w.fr "Hintergrund-Relief" $titlerelief $titlebd
    createReliefItem $w.fr $cterm flat flach
    createReliefItem $w.fr $cterm sunken eingesunken
    createReliefItem $w.fr $cterm raised gehoben

    frame $w.fb -class TConfigFrame
    button $w.fb.ok -class TConfigButton -text "OK" \
          -command "doneAtts $w $term $cterm"  -width 10
    button $w.fb.doit -class TConfigButton -text "Anwenden" \
          -command "doAtts $w $term $cterm" -width 10
    button $w.fb.cancel -class TConfigButton -text "Abbrechen" \
          -command "cancelAtts $w $term $cterm" -width 10
    pack  $w.fb.ok $w.fb.doit $w.fb.cancel -side left -fill x -expand 1
    pack  $w.fb -side bottom -fill x -expand 1
    pack $w.fa $w.fc $w.fr -side left -fill both -expand 1 -anchor n
    wm title $w "Attribute"
  } else {
     raise $w $root
  }
}

#creation of a radiobutton
proc createRadioItem { parent name text variable} {
  radiobutton $parent.$name -class TConfigRadiobutton \
     -text $text \
     -variable $variable \
     -anchor w
  pack $parent.$name -fill x
  return $parent.$name
}
#creation of a label and an entry inside a frame
proc createLabelEntryItem { parent name text variable } {
  if { [info globals tkOrgLabels] != "" } {
    global tkOrgLabels
    set oldlabel $tkOrgLabels
    set tkOrgLabels 1
  }
  if { [info globals tkOrgEntrys] != "" } {
    global tkOrgEntrys
    set oldentry $tkOrgEntrys
    set tkOrgEntrys 1
  }
  frame $parent.$name -class TConfigFrame
  #label
  label $parent.$name.l -class TConfigLabel -text $text

  #entry
  entry $parent.$name.e -class TConfigEntry -textvar $variable  \
     -bd 1 -relief sunken
  pack $parent.$name.l -side left
  pack $parent.$name.e -side right
  pack $parent.$name -anchor w  -expand 1 -fill both
  if { [info globals tkOrgLabels] != "" } {
    set tkOrgLabels $oldlabel
  }
  if { [info globals tkOrgEntrys] != "" } {
    set tkOrgEntrys $oldentry
  }
  return $parent.$name
}
proc isnumber { name } {
  upvar $name str
  if { [catch "expr round($str)" msg] == 1 } {
    return 0
  }
  set str $msg
  return 1
}
proc changeConfigSpecs { term } {
  #read in some keys
  global tshowspec_key_var
  switch_keys $tshowspec_key_var

  #set wrap
  global tshowspec_wrap_var
  $term wrap $tshowspec_wrap_var
  #configure width and height
  global tshowspec_height_var
  if { [isnumber tshowspec_height_var] && ( $tshowspec_height_var >= 0 ) } {
    $term configure -height $tshowspec_height_var
  }
  set tshowspec_height_var [lindex [$term configure -height] 4]
  $term configure -height $tshowspec_height_var
  global tshowspec_width_var
  if { [isnumber tshowspec_width_var] && ( $tshowspec_width_var >= 0 ) } {
    $term configure -width $tshowspec_width_var
  }
  set tshowspec_width_var [lindex [$term configure -width] 4]
}

proc cancelSpecs { w term } {
#delete all global variables
  global tshowspec_key_var
  unset tshowspec_key_var
  global tshowspec_width_var
  unset tshowspec_width_var
  global tshowspec_height_var
  unset tshowspec_height_var
  global tshowspec_wrap_var
  unset tshowspec_wrap_var
  destroy $w
  focus $term
}
proc doneSpecs { w term } {
  changeConfigSpecs $term
  cancelSpecs $w $term
}
proc doSpecs { w term } {
  changeConfigSpecs $term
}
#===========================================================
# showSpecs : Displays a Configuration-Dialog for Terminal
# Terminal-Keys,Size,Wrap
#-----------------------------------------------------------

proc showSpecs { root term background} {
  set w $root.l_specs
  if {! [winfo exists $w]} {
    toplevel $w -class TConfigDialog -relief groove -relief raised -bd 1
    set titlerelief flat
    set titlebd 1
    #the keys
    frame $w.fk -class TConfigFrame -bd $titlebd -relief raised
    createTitle     $w.fk "Tasten" $titlerelief $titlebd
    createRadioItem $w.fk vt100 "VT100" tshowspec_key_var
    createRadioItem $w.fk ansi  "ANSI"  tshowspec_key_var
    createRadioItem $w.fk xterm "XTERM" tshowspec_key_var
    createRadioItem $w.fk user  "USER"  tshowspec_key_var
    global wtkeytab
    global tshowspec_key_var
    set tshowspec_key_var $wtkeytab(TYPE)
    #size
    frame $w.fg -class TConfigFrame -bd $titlebd -relief raised
    createTitle     $w.fg "Gr\xf6\xdf\e" $titlerelief $titlebd
    createLabelEntryItem $w.fg width "Spalten" tshowspec_width_var
    createLabelEntryItem $w.fg height "Zeilen" tshowspec_height_var
    global tshowspec_width_var
    set tshowspec_width_var [lindex [$term configure -width] 4]
    global tshowspec_height_var
    set tshowspec_height_var [lindex [$term configure -height] 4]
    #
    frame $w.fw -class TConfigFrame -bd $titlebd -relief raised
    createTitle     $w.fw "Wrap" $titlerelief $titlebd
    createRadioItem $w.fw 1 "an" tshowspec_wrap_var
    createRadioItem $w.fw 0 "aus" tshowspec_wrap_var
    global tshowspec_wrap_var
    set tshowspec_wrap_var [$term wrap]
    frame $w.fb -class TConfigFrame
    button $w.fb.ok -class TConfigButton -text "OK" \
          -command "doneSpecs $w $term " -width 10
    button $w.fb.doit -class TConfigButton -text "Anwenden" \
          -command "doSpecs $w $term " -width 10
    button $w.fb.cancel -class TConfigButton -text "Abbrechen" \
          -command "cancelSpecs $w $term " -width 10
    pack  $w.fb.ok $w.fb.doit $w.fb.cancel -side left -fill x -expand 1
    pack  $w.fb -side bottom -fill x -expand 1
    pack $w.fk $w.fw $w.fg -side left -fill both -expand 1 -anchor n
    wm title $w "Terminal-Einstellungen"
  } else {
     raise $w $root
  }
}

#===========================================================
# showThisColor :
# Parameter: root : is the frame witch is the father of
#		terminal.
#-----------------------------------------------------------

proc showThisColor { root term background} {
global SelectedColor
   set newColorI [ lindex [$root.l_colors.lb curselection] 0]
   if {$newColorI != ""} {
	set newColorS [$root.l_colors.lb get $newColorI]
	set SelectedColor [lindex [$term configure -$newColorS] 4]
	mkcolor .top $SelectedColor "setThisColor $term $newColorS " $background
   } else {
     focus $root
   }
   destroy $root.l_colors
}

#===========================================================
# setThisColor :
# Parameter: root : is the frame witch is the father of
#		terminal.
#-----------------------------------------------------------

proc setThisColor { term whatColor colorCode } {
	$term configure -$whatColor $colorCode
}



