#include <stdlib.h>
#include <tk.h>
#include <X11/Xproto.h>

#define BI_RGB  0
#define BI_RLE8 1
#define BI_RLE4 2

#define MAX_COLOR 256
#define MAX_ALLOCCOLOR 100
#define CLOSED_COLOR(dist) (dist < 255*3 )
#define ABS(x) ((x)>0?(x):-(x))

typedef unsigned char byte;

#define WIN_OS2_OLD 12
#define WIN_NEW     40
#define OS2_NEW     64
#undef NULL
#define NULL 0

static long filesize;

static char *LoadBMP   _ANSI_ARGS_((char *, int *, int *,int *,XColor **,long *, char *));
static int   loadBMP1   _ANSI_ARGS_((FILE *, byte *, unsigned int, unsigned int));
static int   loadBMP4   _ANSI_ARGS_((FILE *, byte *, unsigned int, unsigned int, unsigned int));
static int   loadBMP8   _ANSI_ARGS_((FILE *, byte *, unsigned int, unsigned int, unsigned int));
static unsigned int getshortBmp   _ANSI_ARGS_((FILE *));
static unsigned int getintBmp     _ANSI_ARGS_((FILE *));
static void PutNearColor(Tk_Window ,unsigned char *,int ,int,XColor *,long);


#define FERROR(fp) (ferror(fp) || feof(fp))


int TkReadBMPFile(Tk_Window tkwin,char *string,int *width,int *height,
                  Pixmap *pixmap,char *readError) {
static XGCValues gc_values;
XImage *xImage;
GC bitmap_gc;
int i,j;
/*u_char * imageMemory;*/
char * imageMemory;
unsigned int bitCount;
XColor *colorbmp;
long nbcolor;

  if ((imageMemory = LoadBMP(string,width,height,&bitCount,&colorbmp,&nbcolor,readError)) == NULL)
     return 0;

  PutNearColor(tkwin,(unsigned char *)imageMemory,*width,*height,colorbmp,nbcolor);

/*printf("here2 %s, %d,%d, %d,(%s)\n",string,*width,*height,bitCount,readError);*/
  bitCount=8;
  xImage = XCreateImage(Tk_Display(tkwin),Tk_Visual(tkwin),bitCount,ZPixmap,0,(char *)imageMemory,(*width),(*height),8,(*width)*bitCount/8);
  /*xImage->bitmap_bit_order = MSBFirst;*/
  /*xImage->byte_order = MSBFirst;*/
  *pixmap = XCreatePixmap(Tk_Display(tkwin), RootWindowOfScreen(Tk_Screen(tkwin)), xImage->width, xImage->height, bitCount);
  gc_values.function = GXcopy;
  gc_values.plane_mask = AllPlanes;
  gc_values.foreground = 1;
  gc_values.background = 0;

  bitmap_gc = XCreateGC(Tk_Display(tkwin), *pixmap, GCForeground | GCBackground, &gc_values);
  XPutImage(Tk_Display(tkwin), *pixmap, bitmap_gc, xImage, 0, 0, 0, 0, xImage->width, xImage->height);

  XFreeGC(Tk_Display(tkwin),bitmap_gc);

  XDestroyImage(xImage);
  return 1;
}

char * LoadBMP(fname,width,height,bitCount,colorbmp,nbcolor,readError )
char *fname;
int  *width;
int *height;
int  *bitCount;
XColor **colorbmp;
long *nbcolor;
char *readError;
{
  FILE         *fp;
  int          i, c, c1, rv;
  unsigned int bfSize, bfOffBits, biSize, biWidth, biHeight, biPlanes;
  unsigned int biBitCount, biCompression, biSizeImage, biXPelsPerMeter;
  unsigned int biYPelsPerMeter, biClrUsed, biClrImportant;
  int bPad;
  char *pic8;
  int cmaplen;

  /* returns pic8 on success or NULL if error */

  pic8 = (byte *) NULL;

  fp = fopen(fname,"r");
  if (!fp) { strcpy(readError,"couldn't open file"); return NULL; }
  
  fseek(fp, 0L, 2);      /* figure out the file size */
  filesize = ftell(fp);
  fseek(fp, 0L, 0);


  /* read the file type (first two bytes) */
  c = getc(fp);  c1 = getc(fp);
  if (c!='B' || c1!='M') { strcpy(readError,"file type != 'BM'"); goto ERROR; }

  bfSize = getintBmp(fp);
  getshortBmp(fp);         /* reserved and ignored */
  getshortBmp(fp);
  bfOffBits = getintBmp(fp);

  biSize          = getintBmp(fp);

  if (biSize == WIN_NEW || biSize == OS2_NEW) {
    biWidth         = getintBmp(fp);
    biHeight        = getintBmp(fp);
    biPlanes        = getshortBmp(fp);
    biBitCount      = getshortBmp(fp);
    biCompression   = getintBmp(fp);
    biSizeImage     = getintBmp(fp);
    biXPelsPerMeter = getintBmp(fp);
    biYPelsPerMeter = getintBmp(fp);
    biClrUsed       = getintBmp(fp);
    biClrImportant  = getintBmp(fp);
  }

  else {    /* old bitmap format */
    biWidth         = getshortBmp(fp);          /* Types have changed ! */
    biHeight        = getshortBmp(fp);
    biPlanes        = getshortBmp(fp);
    biBitCount      = getshortBmp(fp);
    
    /* Not in old versions so have to compute them*/
    biSizeImage = (((biPlanes * biBitCount*biWidth)+31)/32)*4*biHeight;
    
    biCompression   = BI_RGB; 
    biXPelsPerMeter = biYPelsPerMeter = 0;
    biClrUsed       = biClrImportant  = 0;
  }

  if (FERROR(fp)) { strcpy(readError,"EOF reached in file header"); goto ERROR; }

  if ( biBitCount >8) {
    strcpy(readError,"more than 8 bits colors not supported!");
    goto ERROR;
  }

  /* error checking */
  if ((biBitCount!=1 && biBitCount!=4 && biBitCount!=8 ) || 
      biPlanes!=1 || biCompression>BI_RLE4) {

    sprintf(readError,"Bogus BMP File!  (bitCount=%d, Planes=%d, Compression=%d)",
	    biBitCount, biPlanes, biCompression);

    goto ERROR;
  }

  if (((biBitCount==1 ) && biCompression != BI_RGB) ||
      (biBitCount==4 && biCompression==BI_RLE8) ||
      (biBitCount==8 && biCompression==BI_RLE4)) {

    sprintf(readError,"Bogus BMP File!  (bitCount=%d, Compression=%d)",
	    biBitCount, biCompression);

    goto ERROR;
  }


  bPad = 0;
  if (biSize != WIN_OS2_OLD) {
    /* skip ahead to colormap, using biSize */
    c = biSize - 40;    /* 40 bytes read from biSize to biClrImportant */
    for (i=0; i<c; i++) getc(fp);
    
    bPad = bfOffBits - (biSize + 14);
  }

  /* load up colormap, if any */

  nbcolor[0] = (biClrUsed) ? biClrUsed : 1 << biBitCount;
  if ( nbcolor[0] )
    colorbmp[0] = (XColor *)malloc(sizeof(XColor)*nbcolor[0]);
  for (i=0; i<nbcolor[0]; i++) {
    /*printf("cmaplen=%d\n",cmaplen);
      printf("second=%d\n",getc(fp));
      printf("second=%d\n",getc(fp));
      printf("third=%d\n",getc(fp));*/
    colorbmp[0][i].blue  = getc(fp);
    colorbmp[0][i].green = getc(fp);
    colorbmp[0][i].red   = getc(fp);
    if (biSize != WIN_OS2_OLD) {
      getc(fp);
      bPad -= 4;
    }
  }

  if (FERROR(fp)) {
    strcpy(readError,"EOF reached in BMP colormap"); 
    goto ERROR;
   }

  if (biSize != WIN_OS2_OLD) {
    /* Waste any unused bytes between the colour map (if present)
       and the start of the actual bitmap data. */
    
    while (bPad > 0) {
      (void) getc(fp);
      bPad--;
    }
  }

  /* create pic8 */

  pic8 = (byte *) calloc((size_t) biWidth * biHeight, (size_t) 1);
  if (!pic8) {strcpy(readError, "couldn't malloc 'pic8'"); return NULL;}

  /* load up the image */
  if      (biBitCount == 1) rv = loadBMP1(fp,pic8,biWidth,biHeight);
  else if (biBitCount == 4) rv = loadBMP4(fp,pic8,biWidth,biHeight,
					  biCompression);
  else if (biBitCount == 8) rv = loadBMP8(fp,pic8,biWidth,biHeight,
					  biCompression);

  if (rv) { strcpy(readError, "File appears truncated.  Winging it.\n");
             return NULL;}

  fclose(fp);


  *width  = biWidth;
  *height = biHeight;
  *bitCount = biBitCount;

/*
  fprintf(stdout,"width=%d,height=%d\n",biWidth,biHeight);
  c=0;
  for ( rv = 0; rv < biHeight; rv++ ) {
    for ( i = 0; i < biWidth; i++ )
       fprintf(stdout,"%c",pic8[c++]+32);
    fprintf(stdout,"\n"); fflush(stdout);
  }
*/

  return pic8;


 ERROR:
  fclose(fp);
  return NULL;
}


/*******************************************/
static int loadBMP1(fp, pic8, w, h)
     FILE *fp;
     byte *pic8;
     unsigned int  w,h;
{
  int   i,j,c,bitnum,padw;
  byte *pp;

  c = 0;
  padw = ((w + 31)/32) * 32;  /* 'w', padded to be a multiple of 32 */

  for (i=h-1; i>=0; i--) {
    pp = pic8 + (i * w);
    /*if ((i&0x3f)==0) WaitCursor();*/
    for (j=bitnum=0; j<padw; j++,bitnum++) {
      if ((bitnum&7) == 0) { /* read the next byte */
	c = getc(fp);
	bitnum = 0;
      }
      
      if (j<w) {
	*pp++ = (c & 0x80) ? 1 : 0;
	c <<= 1;
      }
    }
    if (FERROR(fp)) break;
  }

  return (FERROR(fp));
}  



/*******************************************/
static int loadBMP4(fp, pic8, w, h, comp)
     FILE *fp;
     byte *pic8;
     unsigned int  w,h,comp;
{
  int   i,j,c,c1,x,y,nybnum,padw,rv;
  byte *pp;
  
  
  rv = 0;
  c = c1 = 0;
  
  if (comp == BI_RGB) {   /* read uncompressed data */
    padw = ((w + 7)/8) * 8; /* 'w' padded to a multiple of 8pix (32 bits) */
    
    for (i=h-1; i>=0; i--) {
      pp = pic8 + (i * w);
      /*if ((i&0x3f)==0) WaitCursor();*/
      
      for (j=nybnum=0; j<padw; j++,nybnum++) {
	if ((nybnum & 1) == 0) { /* read next byte */
	  c = getc(fp);
	  nybnum = 0;
	}
	
	if (j<w) {
	  *pp++ = (c & 0xf0) >> 4;
	  c <<= 4;
	}
      }
      if (FERROR(fp)) break;
    }
  }
  
  else if (comp == BI_RLE4) {  /* read RLE4 compressed data */
    x = y = 0;  
    pp = pic8 + x + (h-y-1)*w;
    
    while (y<h) {
      c = getc(fp);  if (c == EOF) { rv = 1;  break; }
      
      if (c) {                                   /* encoded mode */
	c1 = getc(fp);
	for (i=0; i<c; i++,x++,pp++) 
	  *pp = (i&1) ? (c1 & 0x0f) : ((c1>>4)&0x0f);
      }
      
      else {    /* c==0x00  :  escape codes */
	c = getc(fp);  if (c == EOF) { rv = 1;  break; }
	
	if      (c == 0x00) {                    /* end of line */
	  x=0;  y++;  pp = pic8 + x + (h-y-1)*w;
	} 
	
	else if (c == 0x01) break;               /* end of pic8 */
	
	else if (c == 0x02) {                    /* delta */
	  c = getc(fp);  x += c;
	  c = getc(fp);  y += c;
	  pp = pic8 + x + (h-y-1)*w;
	}
	
	else {                                   /* absolute mode */
	  for (i=0; i<c; i++, x++, pp++) {
	    if ((i&1) == 0) c1 = getc(fp);
	    *pp = (i&1) ? (c1 & 0x0f) : ((c1>>4)&0x0f);
	  }
	  
	  if (((c&3)==1) || ((c&3)==2)) getc(fp);  /* read pad byte */
	}
      }  /* escape processing */
      if (FERROR(fp)) break;
    }  /* while */
  }
  
  else {
    fprintf(stderr,"unknown BMP compression type 0x%0x\n", comp);
  }
  
  if (FERROR(fp)) rv = 1;
  return rv;
}  



/*******************************************/
static int loadBMP8(fp, pic8, w, h, comp)
     FILE *fp;
     byte *pic8;
     unsigned int  w,h,comp;
{
  int   i,j,c,c1,padw,x,y,rv;
  byte *pp;
  
  rv = 0;

  if (comp == BI_RGB) {   /* read uncompressed data */
    padw = ((w + 3)/4) * 4; /* 'w' padded to a multiple of 4pix (32 bits) */

    for (i=h-1; i>=0; i--) {
      pp = pic8 + (i * w);
      /*if ((i&0x3f)==0) WaitCursor();*/

      for (j=0; j<padw; j++) {
	c = getc(fp);  if (c==EOF) rv = 1;
	if (j<w) *pp++ = c;
      }
      if (FERROR(fp)) break;
    }
  }

  else if (comp == BI_RLE8) {  /* read RLE8 compressed data */
    x = y = 0;  
    pp = pic8 + x + (h-y-1)*w;

    while (y<h) {
      c = getc(fp);  if (c == EOF) { rv = 1;  break; }

      if (c) {                                   /* encoded mode */
	c1 = getc(fp);
	for (i=0; i<c; i++,x++,pp++) *pp = c1;
      }

      else {    /* c==0x00  :  escape codes */
	c = getc(fp);  if (c == EOF) { rv = 1;  break; }

	if      (c == 0x00) {                    /* end of line */
	  x=0;  y++;  pp = pic8 + x + (h-y-1)*w;
	} 

	else if (c == 0x01) break;               /* end of pic8 */

	else if (c == 0x02) {                    /* delta */
	  c = getc(fp);  x += c;
	  c = getc(fp);  y += c;
	  pp = pic8 + x + (h-y-1)*w;
	}

	else {                                   /* absolute mode */
	  for (i=0; i<c; i++, x++, pp++) {
	    c1 = getc(fp);
	    *pp = c1;
	  }
	  
	  if (c & 1) getc(fp);  /* odd length run: read an extra pad byte */
	}
      }  /* escape processing */
      if (FERROR(fp)) break;
    }  /* while */
  }
  
  else {
    fprintf(stderr,"unknown BMP compression type 0x%0x\n", comp);
  }

  if (FERROR(fp)) rv = 1;
  return rv;
}  


/*******************************************/
static unsigned int getshortBmp(fp)
     FILE *fp;
{
  int c, c1;
  c = getc(fp);  c1 = getc(fp);
  return ((unsigned int) c) + (((unsigned int) c1) << 8);
}


/*******************************************/
static unsigned int getintBmp(fp)
     FILE *fp;
{
  int c, c1, c2, c3;
  c = getc(fp);  c1 = getc(fp);  c2 = getc(fp);  c3 = getc(fp);
  return ((unsigned int) c) +
         (((unsigned int) c1) << 8) + 
	 (((unsigned int) c2) << 16) +
	 (((unsigned int) c3) << 24);
}
/*******************************************************************************
** FUNCTION PutNearColor : Find near color of a BMP image in the current
**                         ColorMap
** Parameters :
**  tkwin    : TK Windows
**  data     : Date to process. Each byte correspond to one pixel
**  width    : Width of data
**  height   : Height of data
**  colorbmp : Current color map of the data
**  nbcolor  : Number of color in data
** Returnings :
**  none
*******************************************************************************/

void PutNearColor(tkwin,data,width,height,colorbmp,nbcolor)

Tk_Window tkwin;
unsigned char *data;
int width;
int height;
XColor *colorbmp;
long nbcolor;

{
 int nbExistColor=0;	   /* Number of existing color in current colormap*/
 XColor currCol;	   /* work variable */
 XColor tabCol[MAX_COLOR]; /* Color of bmp image */
 long usedColor[MAX_COLOR];/* Number used for a bmp color */
 long indColor[MAX_COLOR]; /* Index of bmp used color in descending modus */
 register long i,j;	   /* work variables */
 char multiply=1;
 long mindist;
 long maxUsed;
 long currdist;
 long ind;
 char found;
 char moreColor;
 long iMax;

/****************************************************
* Count how many colors is in the current color map *
****************************************************/
 currCol.pixel=0;
 currCol.blue=1;

 while ( nbExistColor < MAX_COLOR
   && (currCol.blue != 0 || currCol.red != 0 || currCol.green != 0) ) {
   XQueryColor(Tk_Display(tkwin),Tk_Colormap(tkwin),&currCol);
   memcpy(&tabCol[nbExistColor],&currCol,sizeof(XColor));
   currCol.pixel++;
   nbExistColor++;
  }

 currCol.blue=1;
 if ( nbExistColor == 2 )
   while ( nbExistColor < MAX_COLOR
   && (currCol.blue != 0 || currCol.red != 0 || currCol.green != 0) ) {
     XQueryColor(Tk_Display(tkwin),Tk_Colormap(tkwin),&currCol);
     memcpy(&tabCol[nbExistColor],&currCol,sizeof(XColor));
     currCol.pixel++;
     nbExistColor++;
    }

/***********************************************************
* If value of bmp color rgb < 256 -> put it in right modus *
***********************************************************/

 for (i=0; i < nbcolor; i++) {
   usedColor[i]=0;
   indColor[i]=i;
   if ( colorbmp[i].red > 255 || colorbmp[i].green > 255
    || colorbmp[i].blue > 255) {
     multiply=0;
     break;
   }
 }

 if ( multiply ) {
   for (i=0; i < nbcolor; i++) {
     colorbmp[i].red   <<=8;
     colorbmp[i].green <<= 8;
     colorbmp[i].blue  <<= 8;
   }
 }

 moreColor = ( nbExistColor < MAX_ALLOCCOLOR );
 if ( moreColor ) {
   /************************************
   * Count for each color how many use *
   ************************************/
   ind=0;
   for ( i = 0; i < height; i++ ) {
     for ( j = 0; j < width; j++ )
        usedColor[data[ind++]]++;
   }
   /*printf("UsedColor=(");
   for ( i = 0; i < nbcolor; i++ ) {
     printf("%d,",usedColor[i]);
   }
   printf(")\n");*/
   /**************************
   * Sort in descending mode *
   **************************/
   for ( i = 0; i < nbcolor; i++ ) {
     for ( j = i+1; j < nbcolor; j++ ) {
       if ( usedColor[i] < usedColor[j] ) {
         maxUsed = usedColor[i];
         usedColor[i] = usedColor[j];
         usedColor[j] = maxUsed;
         ind = indColor[i];
         indColor[i] = indColor[j];
         indColor[j] = ind;
       }
     }
   }
 }
/*
 printf("indColor=(");
 for ( i = 0; i < nbcolor; i++ ) {
   printf("%d,",indColor[i]);
 }
 printf(")\n");
*/

/************************************************
* Search the closed color for each color of BMP *
************************************************/
 for (i=0; i < nbcolor; i++) {
   mindist=0;
   ind=-1;
   found=0;
   iMax=indColor[i];
   for (j=0; j < nbExistColor; j++) {
     currdist =  ABS(tabCol[j].red   - colorbmp[iMax].red)
               + ABS(tabCol[j].green - colorbmp[iMax].green)
               + ABS(tabCol[j].blue  - colorbmp[iMax].blue);
     if ( ind == (-1) || currdist < mindist ) {
       mindist = currdist;
       ind = j;
       if ( CLOSED_COLOR(mindist) ) {
         found = 1;
         break;
       }
     }
   }
   /****************************************************
   * If we don't find a near color and we have enough  *
   * color to allocate, try to allocate a new one      *
   ****************************************************/
   if ( moreColor && !found && nbExistColor < MAX_ALLOCCOLOR ) {
     colorbmp[iMax].flags = DoRed|DoGreen|DoBlue;
     if (XAllocColor(Tk_Display(tkwin),Tk_Colormap(tkwin),&colorbmp[iMax]) == 0) {
       moreColor=0;
     } else {
       memcpy(&tabCol[nbExistColor],&colorbmp[iMax],sizeof(XColor));
       ind = nbExistColor;
       nbExistColor++;
     }
   }
   colorbmp[iMax].pixel = tabCol[ind].pixel;
 }

 /******************************************
 * "Correct" data with the right pixel map *
 ******************************************/
 ind=0;
 for ( i = 0; i < height; i++ ) {
   for ( j = 0; j < width; j++ ) {
     data[ind]=colorbmp[data[ind++]].pixel;
   }
 }

/*
 printf("NbColor=%d\n",nbExistColor);

 for (j=0; j < nbcolor; j++) {
   printf("B[%03d-%04x-%04x-%04x]",colorbmp[j].pixel,colorbmp[j].red,colorbmp[j].green,colorbmp[j].blue);
   for (i=0; i < nbExistColor; i++)
     if ( colorbmp[j].pixel == tabCol[i].pixel ) 
        break;
   printf("-C[%03d-%04x-%04x-%04x]\n",tabCol[i].pixel,tabCol[i].red,tabCol[i].green,tabCol[i].blue);
 }
*/


} /* END FUNCTION PutNearColor */
