#include "tk.h"

#define max(A,B) ((A)>(B))?(A):(B)
#define min(A,B) ((A)>(B))?(B):(A)

#define TBL_REDRAW_PENDING	1	/* redraw event already pending */
#define TBL_CURSOR_ON		2	/* The edit cursor is on */
#define	TBL_HAS_FOCUS		4	/* the window has the focus */
#define TBL_TEXT_CHANGED	8	/* the text has been edited */
#define TBL_EDIT_ALLOWED	16	/* are edits allowed */
#define	TBL_1_KEY_DESTROY	32	/* Does the first key in an edit delete the whole string */
#define TBL_FLASH_ENABLED	64	/* is flashing enabled */
#define	TBL_BATCH_ON		128	/* no updates are forced */

#define	DRAW_MODE_SLOW		1	/* The display redraws with a pixmap using TK function calls */
#define	DRAW_MODE_TK_COMPAT	2	/* The redisplay is direct to the screen, but TK function calls
					** are still used to give correct 3-d border appearance and
					** thus remain compatible with other TK apps */
#define DRAW_MODE_FAST		3	/* the redisplay goes straight to the screen and the
					** 3d borders are rendered with a single pixel wide line
					** only. It cheats and uses the internal border structure
					** to do the borders */

#define	STRETCH_MODE_NONE	1	/* No additional pixels will be added to rows or cols */
#define	STRETCH_MODE_UNSET	2	/* All default rows or columns will be stretched to fill the screen */
#define STRETCH_MODE_ALL	3	/* All rows/columns will be padded to fill the window */

/* ripped  out of the tk3d.c code */
typedef struct {
    Display *display;		
    int refCount;		
    XColor *bgColorPtr;		
    XColor *lightColorPtr;	
    XColor *darkColorPtr;	
    Pixmap shadow;		
    GC lightGC;			
    GC darkGC;			
    GC bgGC;			
    Tcl_HashEntry *hashPtr;	
} Border;


/* The tag structure */
typedef struct {
	/*
	** These are the requested values for
	** the tag, except in the default tag
	** they can be null, in which case values
	** from the default will be used
	*/

	Tk_3DBorder	configBgBorder;		/* border colour structure */
	XColor		*configForeground;	/* foreground colour */
	int		configRelief;		/* relief type */
	XFontStruct	*configFontPtr;		/* default font pointer */
	Tk_Anchor	configAnchor;		/* default anchor point */

	/* 
	** the rest is a cache of the actual values to
	** use for some of the parameters. These are set
	** either from the above values, or from the
	** default tag values 
	*/
	Tk_3DBorder	bgBorder;
	int		relief;		/* relief type */
	XFontStruct	*fontPtr;	/* default font pointer */
	Tk_Anchor	anchor;		/* default anchor point */
	GC		copyGc;		/* used to copy stuff */
} tagStruct ;

/* 
** The widget structure for the table Widget
*/
typedef struct {
	/* 
	** basic information about the 
	** window and the interpreter 
	*/
	Tk_Window 	tkwin;	
	Display 	*display;
	Tcl_Interp	*interp;
	
	/* Configurable things */
	int		rows, cols;	/* number of rows and columns */
	int		defRowHeight;	/* default row height in pixels */
	int		defColWidth;	/* default column width in chars */
	int		maxReqWidth;	/* the maximum requested width in pixels */
	int		maxReqHeight;	/* the maximum requested height in pixels */
	char 		*arrayVar;	/* name af array variable */
	int		borderWidth;	/* internal borderwidth */
	tagStruct	defaultTag;	/* the default tag colours/fonts etc */
	char 		*yScrollCmd;	/* the y-scroll command */
	char		*xScrollCmd;	/* the x-scroll command */
	Tk_3DBorder	cursorBg;	/* the cursor colour */
	int		rowThenCol;	/* is the table in row then col mode */
	int		selectionOn;	/* Is selection enabled */

	/* 
	** the topleft cell displayed excluding the fixed title rows 
	** this is just the configuration request. The actual cell 
	** used may be different to keep the screen full
	*/
	int		topRow;		
	int		leftCol;	
	int		titleRows;	/* the number of rows to use as a title */
	int 		titleCols;	/* the number of columns to use as a title */

	/* Cached Information */
	int		selRow, selCol;	/* the row and column of the selected cell */
	int 		oldTopRow;	/* the top row the last time TableAdjust Params */
	int 		oldLeftCol;	/* was called */
	int		oldSelRow;	/* the previous selected row and column */
	int		oldSelCol;
	int		textCurPosn;	/* The position of the text cursor in the string */
	int		tableFlags;	/* An or'ed combination of flags concerning redraw/cursor etc. */
	int		drawMode;	/* The mode to use when redrawing */
	int		colStretchMode;	/* The way to stretch columns if the window is too large */
	int		rowStretchMode;	/* The way to stretch rows if the window is too large */
	int		maxWidth;	/* max width required in pixels */
	int		maxHeight;	/* max height required in pixels */
	int		charWidth;	/* width of a character in the default font */
	int		*colPixels;	/* Array of the pixel width of each column */
	int		*rowPixels;	/* Array of the pixel height of each row */
	int		*colStarts;	/* Array of start pixels for columns */
	int 		*rowStarts;	/* Array of start pixels for rows */
	int		colOffset;	/* X index of the leftmost column in the display */
	int		rowOffset;	/* Y index of the topmost row in the display */
	int		flashTime;	/* The number of ticks to flash a cell for */
	tagStruct	*flashTag;	/* the tag used for flashing cells */
	Tcl_HashTable	*colWidths;	/* hash table of non default column widths */
	Tcl_HashTable	*rowHeights;	/* hash table of non default row heights */
	Tcl_HashTable	*tagTable;	/* the table for the style tags */
	Tcl_HashTable	*rowStyles;	/* the table for row styles */
	Tcl_HashTable	*colStyles;	/* the table for col styles */
	Tcl_HashTable	*cellStyles;	/* the table for cell styles */
	Tcl_HashTable	*flashCells;	/* the table full of flashing cells */
	Tk_TimerToken	cursorTimer;	/* the timer token for the cursor blinking */
	Tk_TimerToken	flashTimer;	/* the timer token for the cell flashing */
	char 		*selectBuf;	/* the buffer where the selection is kept for editing */
	int		selectBufLen;	/* the current length of the buffer */

	/* The invalid recatangle if there is an update pending */
	int		invalidX,
			invalidY,
			invalidWidth,
			invalidHeight;
} Table ;


/* structure for use in command parsing tables */
typedef struct  {
	char 	*name;
	int	value;
} command_struct;



static Tcl_CmdProc		TableCmd;
static Tk_EventProc 		TableEventProc;
static Tcl_CmdProc		TableWidgetCmd;
static Tk_IdleProc  		TableDisplay;
static Tk_FreeProc  		TableDestroy;
static Tcl_VarTraceProc	        TableVarProc;

static void       TableDimensionCalc(Table *);
static void       TableInvalidate(Table *, int, int, int, int, int);
static void       TableCellCoords(Table *, int, int, int *, int *, int *, int *);
static void       TableGetRange(Table *, int, int, int, int,	int *, int *, int *, int *);
static int        parse_command(Tcl_Interp *, command_struct *,char *);
static void       TablePixelCalc(Table *tablePtr);
static void       TableAdjustParams(Table *);
static void       TableTagConfig(Table *, tagStruct *);
static void       TableWhatCell(Table *, int , int , int *, int *);
static tagStruct *TableNewTag( Table *);
static int        TableTagCommand(Table *, int, char **);
static void       TableCleanupTag(Table *, tagStruct *);
static void       TableCursorEvent(ClientData);
static void       TableMakeArrayIndex(Table *, int, int, char *);
static int        TableParseArrayIndex(Table *, int *, int *, char *);
static int        TableParseStringPosn(Table *, char *, int *);
static void       TableBufLengthen(Table *, int);
static void       TableGetSelection(Table *);
static void       TableConfigCursor(Table *tablePtr);
static void       TableFlashConfigure(Table *, int );
static void       TableFlashEvent(ClientData);
static void       TableAddFlash(Table *, int , int );


#define  CMD_CONFIGURE   1  /* general configure command */
#define  CMD_WIDTH  2  /* (re)set column widths */
#define  CMD_HEIGHT  3  /* (re)set row heights */
#define  CMD_GETWIDTH  4  /* print the width of a column */
#define  CMD_GETHEIGHT  5  /* print the height of a column */
#define  CMD_TAG    6  /* tag command menu */
#define CMD_TOPROW  7  /* change the toprow displayed */
#define CMD_LEFTCOL  8  /* change the leftcol displayed*/
#define CMD_WHATCELL  9  /* get the cell at an (x,y) coord */
#define CMD_WHEREIS  10  /* where is the cell (row,col) */
#define CMD_SETCELL  11  /* set the current cell to row, col */
#define CMD_ICURSOR  12  /* set the insertion cursor */
#define CMD_INSERT  13  /* insert text at any position */
#define  CMD_DELETE  14  /* delete text in the selection */
#define  CMD_REREAD  15  /* reread the current selection */
#define CMD_EDITMODE  16  /* change edit mode */
#define CMD_DRAWMODE  17  /* change the drawing mode */
#define CMD_FLASH  18  /* set up the flash parameters */
#define  CMD_BATCH  19  /* command to turn batching on/off */
#define CMD_ROW_STRETCH  20  /* change the row stretching mode */
#define CMD_COL_STRETCH  21  /* change the row stretching mode */

/* The list of commands for the command parser*/
static command_struct commandos[]=
{
  { "configure"  , CMD_CONFIGURE},
  { "width"  , CMD_WIDTH},
  { "height"  , CMD_HEIGHT},
  { "getwidth"  , CMD_GETWIDTH},
  { "getheight"  , CMD_GETHEIGHT},
  { "tag"    , CMD_TAG},
  { "toprow"  , CMD_TOPROW},
  { "leftcol"  , CMD_LEFTCOL},
  { "whatcell"  , CMD_WHATCELL},
  { "whereis"  , CMD_WHEREIS},
  { "setcell"  , CMD_SETCELL},
  { "icursor"  , CMD_ICURSOR},
  { "insert"  , CMD_INSERT},
  { "delete"  , CMD_DELETE},
  { "reread"  , CMD_REREAD},
  { "editmode"  , CMD_EDITMODE},
  { "drawmode"  , CMD_DRAWMODE},
  { "flash"  , CMD_FLASH},
  { "batch"  , CMD_BATCH},
  { "rowstretch"  , CMD_ROW_STRETCH},
  { "colstretch"  , CMD_COL_STRETCH},
  { ""    , 0}
};

/* List of tag subcommands */
#define  TAG_NAMES  1  /* print the tag names */
#define  TAG_DELETE  2  /* delete a tag */
#define  TAG_CONFIGURE  3  /* config/create a new tag */
#define TAG_ROWTAG  4  /* tag a row */
#define TAG_COLTAG  5  /* tag a column */
#define TAG_CELLTAG  6  /* tag a cell */

static command_struct tag_commands[]=
{
  { "names"    , TAG_NAMES },
  { "delete"    , TAG_DELETE },
  { "configure"    , TAG_CONFIGURE },
  { "rowtag"    , TAG_ROWTAG },
  { "coltag"    , TAG_COLTAG },
  { "celltag"    , TAG_CELLTAG },
  { ""       , 0 }
};


/* List of edit subcommands */
#define  EDIT_NOEDIT  1  /* editing is disbaled */
#define EDIT_AUTOCLEAR  2  /* selection is cleared when a key is pressed */
#define  EDIT_NOCLEAR  3  /* no clearing of the selection on first keypress */


static command_struct editmode_commands[]=
{
  { "off"      , EDIT_NOEDIT },
  { "autoclear"    , EDIT_AUTOCLEAR },
  { "noclear"    , EDIT_NOCLEAR },
  { ""      , 0 }
};



/* drawmode subcommands */
static command_struct drawmode_commands[]=
{
  { "fast"    , DRAW_MODE_FAST },
  { "compatible"    , DRAW_MODE_TK_COMPAT},
  { "slow"    , DRAW_MODE_SLOW},
  { ""      , 0 }
};


/* flash subcommands */
#define FLASH_MODE  1
#define  FLASH_TAG  2
#define  FLASH_TIMEOUT  3

static command_struct flash_commands[]=
{
  { "mode"    , FLASH_MODE },
  { "tag"      , FLASH_TAG },
  { "timeout"    , FLASH_TIMEOUT},
  { ""      , 0 }
};


/* stretch subcommands */

static command_struct stretch_commands[]=
{
  { "none"    , STRETCH_MODE_NONE },
  { "unset"    , STRETCH_MODE_UNSET },
  { "all"      , STRETCH_MODE_ALL },
  { ""      , 0 }
};


static Tk_ConfigSpec TableConfig[]= 
{
  {   TK_CONFIG_BORDER, "-background", "background",
    "Background", "BISQUE1",
    Tk_Offset(Table, defaultTag.configBgBorder), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_SYNONYM, "-bg", "background",
    (char *)NULL, (char *)NULL,
    0, 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_COLOR, "-foreground", "foreground",
    "Foreground", "BLACK",
    Tk_Offset(Table, defaultTag.configForeground), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_SYNONYM, "-fg", "foreground",
    (char *)NULL, (char *)NULL,
    0, 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_BORDER, "-cursorbg", "cursor",
    "Cursor", "Black",
    Tk_Offset(Table, cursorBg), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_PIXELS, "-borderwidth", "borderWidth",
    "BorderWidth", "1",
    Tk_Offset(Table, borderWidth), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_INT, "-rows", "rows",
    "Rows", "10",
    Tk_Offset(Table, rows), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_INT, "-cols", "cols",
    "Cols", "10",
    Tk_Offset(Table, cols), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_INT, "-width", "width",
    "Width", "10",
    Tk_Offset(Table, defColWidth), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_INT, "-colorigin", "colorigin",
    "Colorigin", "0",
    Tk_Offset(Table, colOffset), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_PIXELS, "-maxwidth", "maxwidth",
    "Maxwidth", "1000",
    Tk_Offset(Table, maxReqWidth), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_PIXELS, "-maxheight", "maxheight",
    "Maxheight", "800",
    Tk_Offset(Table, maxReqHeight), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_INT, "-roworigin", "roworigin",
    "Roworigin", "0",
    Tk_Offset(Table, rowOffset), 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_INT, "-rowtitle", "rowtitle",
    "Rowtitle", 0,
    Tk_Offset(Table, titleRows), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_INT, "-coltitle", "coltitle",
       "Coltitle", 0,
       Tk_Offset(Table, titleCols), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_INT, "-height", "height",
    "Height", 0,
    Tk_Offset(Table, defRowHeight), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_STRING, "-variable", "variable",
    "Variable", NULL,
    Tk_Offset(Table, arrayVar), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_RELIEF, "-relief", "relief",
    "Relief", "sunken",
    Tk_Offset(Table, defaultTag.configRelief), 0, (Tk_CustomOption *)NULL
  },
  {  TK_CONFIG_FONT, "-font", "font", 
    "Font",  "-Adobe-Helvetica-Bold-R-Normal--*-120-*" ,
    Tk_Offset(Table, defaultTag.configFontPtr), 0, (Tk_CustomOption *)NULL
  },
  {  TK_CONFIG_ANCHOR, "-anchor", "anchor", 
    "Anchor",  "center" ,
    Tk_Offset(Table, defaultTag.configAnchor), 0, (Tk_CustomOption *)NULL
  },  
  {  TK_CONFIG_STRING, "-yscrollcmd", "yscrollcmd", 
    "Yscrollcmd",  NULL ,
    Tk_Offset(Table, yScrollCmd), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {  TK_CONFIG_STRING, "-xscrollcmd", "xscrollcmd", 
    "Xscrollcmd",  NULL ,
    Tk_Offset(Table, xScrollCmd), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {  TK_CONFIG_BOOLEAN, "-rowfirstmode", "rowfirstmode", 
    "Rowfirstmode",  "1" ,
    Tk_Offset(Table, rowThenCol), 0, (Tk_CustomOption *)NULL
  },
  {  TK_CONFIG_BOOLEAN, "-selection", "selection", 
    "Selection", "1", 
    Tk_Offset(Table, selectionOn), 0, (Tk_CustomOption *)NULL
  },
  {  TK_CONFIG_END, (char *)NULL, (char *)NULL, (char *)NULL,
    (char *)NULL, 0, 0
  }
};


/* 
** The default specification for configuring tags
** Done like this to make the command line parsing 
** easy 
*/
static Tk_ConfigSpec tagConfig[] =
{
  {   TK_CONFIG_BORDER, "-background", "background",
    "Background", NULL,
    Tk_Offset(tagStruct, configBgBorder), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_SYNONYM, "-bg", "background",
    (char *)NULL, (char *)NULL,
    0, 0, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_COLOR, "-foreground", "foreground",
    "Foreground", NULL,
    Tk_Offset(tagStruct, configForeground), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {   TK_CONFIG_SYNONYM, "-fg", "foreground",
    (char *)NULL, (char *)NULL,
    0, 0, (Tk_CustomOption *)NULL
  },
  {  TK_CONFIG_FONT, "-font", "font", 
    "Font",  NULL ,
    Tk_Offset(tagStruct, configFontPtr), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {  TK_CONFIG_ANCHOR, "-anchor", "anchor", 
    "Anchor",  NULL ,
    Tk_Offset(tagStruct, configAnchor), 0, (Tk_CustomOption *)NULL
  },  
  {   TK_CONFIG_RELIEF, "-relief", "relief",
    "Relief", NULL,
    Tk_Offset(tagStruct, configRelief), TK_CONFIG_NULL_OK, (Tk_CustomOption *)NULL
  },
  {  TK_CONFIG_END, (char *)NULL, (char *)NULL, (char *)NULL,
    (char *)NULL, 0, 0
  }
};

static int TableConfigure(  Tcl_Interp *interp, Table *tablePtr, int argc, char *argv[], int flags) {
  XGCValues  gcValues;
  GC    newGC;
  char     buf[200];
  Tcl_HashEntry  *hashEntry;
  Tcl_HashSearch  search;
  Tcl_HashEntry *entryPtr;
  tagStruct *tagPtr;
  int dummy, x, y;

  /* 
  ** remove the trace on the old array 
  ** variable if there was one
  */
  if(tablePtr->arrayVar!=NULL)
    Tcl_UntraceVar(  interp, tablePtr->arrayVar, 
         TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
        TableVarProc, (ClientData) tablePtr);

  /* Do the configuration */
  if (Tk_ConfigureWidget(  interp, tablePtr->tkwin, TableConfig, 
        argc, argv, (char *)tablePtr, flags) != TCL_OK)
    return TCL_ERROR;

  /* 
  ** Check the variable exists and
  ** trace it if it does
  */
  if(tablePtr->arrayVar!=NULL)
  {
    /* Create a command for checking the array variable */
    sprintf(buf, "array size %s\n", tablePtr->arrayVar);  

    /* does the variable exist? */
    if(Tcl_Eval(interp, buf)!=TCL_OK)
    {
      /* if not, create it but leave it blank */
      Tcl_SetVar2(interp, tablePtr->arrayVar,  "0,0", "", TCL_GLOBAL_ONLY);
      Tcl_UnsetVar2(interp, tablePtr->arrayVar, "0,0", TCL_GLOBAL_ONLY);
    }

    /* remove the effect of the evaluation */  
    Tcl_ResetResult(interp);

    /* set a trace on the variable */
    Tcl_TraceVar(  interp, tablePtr->arrayVar, 
        TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
        TableVarProc, (ClientData) tablePtr);
  
    /*  get the current value of the selection */
    if(tablePtr->selectionOn)
      TableGetSelection(tablePtr);
  }

  /*
  ** set up the two hash tables for row height and
  ** column width if not already set
  */
  if (tablePtr->colWidths==NULL)
  {
    tablePtr->colWidths=(Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(tablePtr->colWidths,TCL_ONE_WORD_KEYS);
  }
  if (tablePtr->rowHeights==NULL)
  {
    tablePtr->rowHeights=(Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(tablePtr->rowHeights, TCL_ONE_WORD_KEYS);
  }
  /* 
  ** set up the hash tables for style tags 
  ** if it doesn't exist and put the Title
  ** tag into it
  */
  if((tablePtr->tagTable)==NULL)
  {
    char *titleArgs[]={"-bg", "BISQUE2"};
    char *selArgs[]={"-bg", "BISQUE3"};

    tablePtr->tagTable=(Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(tablePtr->tagTable, TCL_STRING_KEYS);
    
    /* Insert the first entry for Titles */
    tagPtr=TableNewTag(tablePtr);


    /* Do a configuration */
    Tk_ConfigureWidget(  interp, tablePtr->tkwin, tagConfig, 
          2, titleArgs, (char *)tagPtr, 0); 
    
    entryPtr=Tcl_CreateHashEntry(tablePtr->tagTable, "Title", &dummy);
    Tcl_SetHashValue(entryPtr, (ClientData)tagPtr);

    /* and an entry for the selected cell */
    tagPtr=TableNewTag(tablePtr);

    /* Do a configuration */
    Tk_ConfigureWidget(  interp, tablePtr->tkwin, tagConfig, 
          2, selArgs, (char *)tagPtr, 0); 
    
    entryPtr=Tcl_CreateHashEntry(tablePtr->tagTable, "sel", &dummy);
    Tcl_SetHashValue(entryPtr, (ClientData)tagPtr);

    /*
    ** now configure the hash table for keeping the
    ** per row, per column, per cell and flashinfo
    */
    tablePtr->rowStyles=(Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(tablePtr->rowStyles, TCL_ONE_WORD_KEYS);

    tablePtr->colStyles=(Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(tablePtr->colStyles, TCL_ONE_WORD_KEYS);

    tablePtr->cellStyles=(Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(tablePtr->cellStyles, TCL_STRING_KEYS);
  
    tablePtr->flashCells=(Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(tablePtr->flashCells, TCL_STRING_KEYS);
  }

  /* Set up the default Tag */
  TableTagConfig(tablePtr, &(tablePtr->defaultTag));

  /*
  ** reset all the other tags too so that 
  ** unspecified options will be set to
  ** the new 'default' ones
  */  
  entryPtr=Tcl_FirstHashEntry(tablePtr->tagTable, &search);
  for(; entryPtr!=NULL; entryPtr=Tcl_NextHashEntry(&search))  
    {  
      char *tagName;

      tagName=Tcl_GetHashKey(tablePtr->tagTable, entryPtr);
      tagPtr=(tagStruct *)Tcl_GetHashValue(entryPtr);
      TableTagConfig(tablePtr, tagPtr); 
    }

  /* set up the default column width and row height */  
  tablePtr->charWidth=XTextWidth(tablePtr->defaultTag.fontPtr, "0", 1);
  if(tablePtr->defRowHeight==0)
    tablePtr->defRowHeight=(tablePtr->defaultTag.fontPtr)->ascent+(tablePtr->defaultTag.fontPtr)->descent+2;


  /* 
  ** Calculate the row and column starts 
  ** Adjust the top left corner of the internal display 
  */
  TableAdjustParams(tablePtr);

  /* reset the cursor */
  TableConfigCursor((ClientData)tablePtr);

  /* set up the background colour in the window*/
  Tk_SetBackgroundFromBorder(tablePtr->tkwin, tablePtr->defaultTag.bgBorder);
  
  /* set the geometry and border */
  x=min(tablePtr->maxWidth,  max( Tk_Width(tablePtr->tkwin), tablePtr->maxReqWidth));
  y=min(tablePtr->maxHeight, max(Tk_Height(tablePtr->tkwin), tablePtr->maxReqHeight));
  Tk_GeometryRequest(tablePtr->tkwin, x, y);
  Tk_SetInternalBorder(tablePtr->tkwin, tablePtr->borderWidth);

  /* invalidate the whole table */
  TableInvalidate(tablePtr, 0, 0, Tk_Width(tablePtr->tkwin), Tk_Height(tablePtr->tkwin),0);

  return TCL_OK;
}


/* 
** The command procedure for a table widget
*/

static int TableWidgetCmd(   ClientData clientdata,
      Tcl_Interp *interp, int argc, char *argv[])
{
  int result, retval, row, col, x, y, x1, y1, subcommand_retval;
  int paramParsed=0;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  Tcl_HashTable  *hashTablePtr;
  int i, column, width, height, dummy, bd, newEntry;
  int entry, key, numfields, maxval, value, posn;
  int posnFrom, posnTo, oldlen, newlen;
  char buf1[50], buf2[50];

  Table *tablePtr=(Table *)clientdata;

  if(argc<2)
  {
    Tcl_AppendResult(  interp, "wrong # args: should be \"",
          argv[0], " option ?arg arg ..?\"",
          (char *)NULL);
    return  TCL_ERROR;
  }
  Tk_Preserve(clientdata);

  /* parse the first parameter */
  retval=parse_command(interp, commandos, argv[1]);

  /* Switch on the parameter value */
  switch(retval)
  {
  /* error */
  case 0:
    /* the return string is already set up */
    result=TCL_ERROR;
    break;

  /* configure */
  case CMD_CONFIGURE: 
    switch (argc) 
    {
    /* Only one argument, print the configuration */
    case 2:
      result=Tk_ConfigureInfo(interp, tablePtr->tkwin, TableConfig,
            (char *)tablePtr, (char *)NULL, 0);
      break;
    /* Two arguments, print the value for one variable */
    case 3:
      result=Tk_ConfigureInfo(interp, tablePtr->tkwin, TableConfig,
            (char *)tablePtr, argv[2], 0);
      break;
    /* More than two arguments, parse the values */
    default:
      result=TableConfigure(  interp, tablePtr, argc-2, argv+2, 
             TK_CONFIG_ARGV_ONLY);
    }
    break;

  /* changes the widths of certain selected columns */
  case CMD_WIDTH:
  case CMD_HEIGHT:
    bd=tablePtr->borderWidth;
    hashTablePtr=(retval==CMD_WIDTH)?(tablePtr->colWidths):(tablePtr->rowHeights);
    maxval=(retval==CMD_WIDTH)?(tablePtr->cols):(tablePtr->rows);

    switch (argc)
    {
    /* print out all the preset column widths or row heights */
    case 2:
      entryPtr=Tcl_FirstHashEntry(hashTablePtr, &search);
      while(entryPtr!=NULL)
      {
        key=(int)Tcl_GetHashKey(hashTablePtr, entryPtr); 
        if(key<maxval)
        {
          sprintf(buf1, "%d", key);
          entry=(int)Tcl_GetHashValue(entryPtr); sprintf(buf2, "%d", entry);
          Tcl_AppendResult(interp, "{",  buf1, " ", buf2, "} ", (char *)NULL);
        }
        entryPtr=Tcl_NextHashEntry(&search);
      }
      result=TCL_OK;  
      break;

    /* add an entry */
    default:
      result=TCL_OK;
      for(i=2;i<argc && result!=TCL_ERROR; i++)
      {
        numfields=sscanf(argv[i], "%d %d", &key, &entry);
        switch(numfields)
        {
        /* no fields, error */
        case 0:
          Tcl_AppendResult(interp, "{" , argv[i], "} invalid pair {", 
              argv[i], "} : usage {int ?int?} ...", (char *)NULL);
          result=TCL_ERROR;
          break;
  
        /* 1 field, reset that value */
        case 1:
          if((entryPtr=Tcl_FindHashEntry(hashTablePtr, (char *)key))!=NULL)
            Tcl_DeleteHashEntry(entryPtr);

          break;
              
        /* 2 fields - set the value */
        case 2:
          /* make sure the value is reasonable */
          entry=max(0,entry);

          entryPtr=Tcl_CreateHashEntry(hashTablePtr, (char *)key, &dummy);
          Tcl_SetHashValue(entryPtr, (ClientData)entry);

          break;
        }
      }
      TableAdjustParams(tablePtr);

      /* rerequest geometry */
      x=min(tablePtr->maxWidth,  max( Tk_Width(tablePtr->tkwin), tablePtr->maxReqWidth));
      y=min(tablePtr->maxHeight, max(Tk_Height(tablePtr->tkwin), tablePtr->maxReqHeight));
      Tk_GeometryRequest(tablePtr->tkwin, x, y);
  
      /*
      ** Invalidate the whole window as TableAdjustParams
      ** will only check to see if the top left cell has moved
      */
      TableInvalidate(tablePtr, 0, 0, Tk_Width(tablePtr->tkwin), Tk_Height(tablePtr->tkwin), 0);
    }
    break;

  /* get the width of a column or the height of a row */
  case CMD_GETWIDTH:
  case CMD_GETHEIGHT:
    /* which hashtable do I need to look at */
    hashTablePtr=(retval==CMD_GETWIDTH)?(tablePtr->colWidths):(tablePtr->rowHeights);

    /* what is the maximum value of the parameter */
    maxval=(retval==CMD_GETWIDTH)?tablePtr->cols:tablePtr->rows;

    if(argc!=3 || sscanf(argv[2], "%d", &key)!=1 || (key<0) || (key>=maxval))
    {
      Tcl_AppendResult(  interp, "usage ", argv[0], 
            (retval==CMD_GETWIDTH) ? " getwidth" : " getheight",
            " arg", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    entryPtr=Tcl_FindHashEntry(hashTablePtr, (char *)key);
    if(entryPtr!=NULL)
      sprintf(buf1, "%d", (int)Tcl_GetHashValue(entryPtr));
    else
    {
      if(retval==CMD_GETWIDTH)
        sprintf(buf1, "%d", tablePtr->defColWidth);
      else
        sprintf(buf1, "%d", tablePtr->defRowHeight);
    }
    Tcl_AppendResult(interp, buf1, (char *)NULL);
    result=TCL_OK;
    break;    

  /* 
  ** Top row or left column to display (after titles)
  ** is being changed. Note that this is relative to
  ** the title, so add back the number of row
  ** and column titles to the value before we
  ** set the internal variables which are absolute column
  ** and row numbers. Usually used by the scrollbars anyway
  */
  case CMD_TOPROW:
  case CMD_LEFTCOL:
    /* just a printout */
    if(argc==2)
    {
      char buf[150];
      if(retval==CMD_TOPROW)
        sprintf(buf, "%d", tablePtr->topRow);
      else
        sprintf(buf, "%d", tablePtr->leftCol);

      Tcl_AppendResult(interp, buf, (char *)NULL);
      result=TCL_OK;
      break;
    }
    if(argc>3 || (sscanf(argv[2], "%d", &value))!=1)
    {
      Tcl_AppendResult(  tablePtr->interp, "wrong # args should be \"",
            argv[0], (retval==CMD_TOPROW)?" toprow":" leftcol", 
            " arg", (char *)NULL);
      result=TCL_ERROR;
      break;
    }

    /* 
    ** set the top row or left column
    ** also adjust the selected cell 
    ** to try to keep it on the screen
    */
    if(retval==CMD_TOPROW)
    {
      tablePtr->topRow=value+tablePtr->titleRows;
      tablePtr->selRow+=(tablePtr->topRow-tablePtr->oldTopRow);
    }
    else
    {
      tablePtr->leftCol=value+tablePtr->titleCols;
      tablePtr->selCol+=(tablePtr->leftCol-tablePtr->oldLeftCol);
    }

    /* 
    ** Do the table adjustment, which will 
    ** also invalidate the table if necessary
    */  
    TableAdjustParams(tablePtr);
    result=TCL_OK;
    break;
    
  /* a veritable plethora of tag commands */
  case CMD_TAG:
  /* do we have another argument */
    if(argc<3)
    {
      Tcl_AppendResult(  tablePtr->interp, "wrong # args should be \"",
            argv[0], " tag option ?arg arg ...?", (char *)NULL);
      result=TCL_ERROR;
      break;
    }

    /* all the rest is now done in a separate function */
    result=TableTagCommand(tablePtr, argc, argv);
    break; /* TAG FUNCTIONS */

  /* get the coords of the cell at the x,y address */
  case CMD_WHATCELL:
    if(argc!=4 || (sscanf(argv[2], "%d", &x)!=1) || (sscanf(argv[3], "%d", &y)!=1))
    {
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], " whatcell arg arg ", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    TableWhatCell(tablePtr, x, y, &row, &col);
    sprintf(buf1, "%d %d", row+tablePtr->rowOffset, col+tablePtr->colOffset);
    Tcl_AppendResult(tablePtr->interp, buf1, (char *)NULL);
    result=TCL_OK;
    break; /* WHAT CELL */

  /* where is the cell (row,col) */
  case CMD_WHEREIS:
    if(argc!=4 || (sscanf(argv[2], "%d", &row)!=1) || (sscanf(argv[3], "%d", &col)!=1))
    {
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], " whereis arg arg ", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    TableCellCoords(tablePtr, row-tablePtr->rowOffset, col-tablePtr->colOffset, &x, &y, &width, &height);
    sprintf(buf1, "%d %d %d %d", x, y, width, height);
    Tcl_AppendResult(tablePtr->interp, buf1, (char *)NULL);
    result=TCL_OK;
    break; /* WHERE IS */
  

  /* set the current cell */
  case CMD_SETCELL:
    /* printout only */
    if(argc==2)
    {
      sprintf(buf1, "%d %d", 
        tablePtr->selRow+tablePtr->rowOffset, 
        tablePtr->selCol+tablePtr->colOffset);
      Tcl_AppendResult(tablePtr->interp, buf1, (char *)NULL);
      result=TCL_OK;  
      break;
    }  
    if(argc!=4 || (sscanf(argv[2], "%d", &row)!=1) || (sscanf(argv[3], "%d", &col)!=1))
    {
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], " setcell arg arg ", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    
    row-=tablePtr->rowOffset; col-=tablePtr->colOffset;
    
    row=min(max(tablePtr->titleRows, row), tablePtr->rows-1);
    col=min(max(tablePtr->titleCols, col), tablePtr->cols-1);

    tablePtr->selRow=row, tablePtr->selCol=col;

    /* Adjust the table for top left, selection on screen etc */
    TableAdjustParams(tablePtr);
    TableConfigCursor(tablePtr);
    result=TCL_OK;
    break;

  /* set the insertion cursor */
  case CMD_ICURSOR:
    switch(argc)
    {
    case 2:
      sprintf(buf1, "%d", tablePtr->textCurPosn);
      Tcl_AppendResult( tablePtr->interp, buf1, (char *)NULL);
      result=TCL_OK;
      break;
    case 3:
      if(TableParseStringPosn(tablePtr, argv[2], &posn)!=TCL_OK)
      {
        result=TCL_ERROR;
        break;
      }
      tablePtr->textCurPosn=posn;
      TableCellCoords(tablePtr, tablePtr->selRow, tablePtr->selCol, &x, &y, &width, &height);
      TableInvalidate(tablePtr, x, y, width, height, 1);
      result=TCL_OK;
      break;
    default:
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], "\" icursor arg", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    break;

  case CMD_INSERT:
    /* are edits enabled */
    if(!(tablePtr->tableFlags&TBL_EDIT_ALLOWED))      
    {
      result=TCL_OK;
      break;
    }
    if(argc!=4)
    {
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], "\" insert index text", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    if(TableParseStringPosn(tablePtr, argv[2], &posn)!=TCL_OK)
    {
      result=TCL_ERROR;
      break;
    }
    
    /* Is this an autoclear and this is the first update */
    if((tablePtr->tableFlags&TBL_1_KEY_DESTROY)&&!(tablePtr->tableFlags&TBL_TEXT_CHANGED))
    {
      /* set the buffer to be empty */
      tablePtr->selectBuf[0]='\0';

      /* mark the text as changed*/
      tablePtr->tableFlags|=TBL_TEXT_CHANGED;

      /* the insert position now has to be 0 */
      posn=0;
    }

    /* get the buffer to at least the right length */
    TableBufLengthen(tablePtr, strlen(tablePtr->selectBuf)+strlen(argv[3]));
    
    /* cache the old string length and the new*/
    oldlen=strlen(tablePtr->selectBuf); newlen=strlen(argv[3]);
    for(i=oldlen; i>=posn;i--)
      tablePtr->selectBuf[i+newlen]=tablePtr->selectBuf[i];
    for(i=0; i<newlen ; i++)
      tablePtr->selectBuf[posn+i]=argv[3][i];

    /* put the cursor after the insertion */
    tablePtr->textCurPosn=posn+newlen;

    TableCellCoords(tablePtr, tablePtr->selRow, tablePtr->selCol, &x, &y, &width, &height);
    TableInvalidate(tablePtr, x, y, width, height, 1);
    result=TCL_OK;
    break;

  case CMD_DELETE:
    /* are edits even enabled */
    if(!(tablePtr->tableFlags&TBL_EDIT_ALLOWED))      
    {
      result=TCL_OK;
      break;
    }

    if(argc>4)
    {
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], "\" delete index ?index?", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    if(  (TableParseStringPosn(tablePtr, argv[2], &posnFrom)!=TCL_OK) ||
      ((argc==4) && TableParseStringPosn(tablePtr, argv[3], &posnTo)!=TCL_OK))
    {
      result=TCL_ERROR;
      break;
    }
    if(argc==3)
      posnTo=posnFrom;
    else if(posnTo<posnFrom)
    {
      int tmp;
      tmp=posnFrom;
      posnFrom=posnTo;
      posnTo=tmp;
    }

    /* do the deletion */
    oldlen=strlen(tablePtr->selectBuf);
    for(i=0; i<=oldlen-posnTo; i++)
      tablePtr->selectBuf[i+posnFrom]=tablePtr->selectBuf[i+posnTo+1];

    /* make sure this string is null terminated */
    tablePtr->selectBuf[oldlen-(posnTo-posnFrom)]='\0';

    /* put the cursor at the deletion point */
    tablePtr->textCurPosn=posnTo;
  
    /* mark the text as changed */
    tablePtr->tableFlags|=TBL_TEXT_CHANGED;

    TableCellCoords(tablePtr, tablePtr->selRow, tablePtr->selCol, &x, &y, &width, &height);
    TableInvalidate(tablePtr, x, y, width, height,1);
    result=TCL_OK;
    break;

  /* this rereads the selection from the array - useful when <escape> is pressed */
  case CMD_REREAD:
    TableGetSelection(tablePtr);
    TableCellCoords(tablePtr, tablePtr->selRow, tablePtr->selCol, &x, &y, &width, &height);
    TableInvalidate(tablePtr, x, y, width, height,1);
    result=TCL_OK;
    break;
  
  /* change the edit mode */
  case CMD_EDITMODE:
    if(argc!=3)
    {
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], "\" edit editmode", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    subcommand_retval=parse_command(tablePtr->interp, editmode_commands, argv[2]);
    switch(subcommand_retval)
    {
    /* not parsed break */
    case 0:
      result=TCL_ERROR;
      break;

    case EDIT_NOEDIT:
      tablePtr->tableFlags&=~TBL_EDIT_ALLOWED;
      break;
    case EDIT_AUTOCLEAR:
      tablePtr->tableFlags|=TBL_EDIT_ALLOWED;
      tablePtr->tableFlags|=TBL_1_KEY_DESTROY;
      break;
    case EDIT_NOCLEAR:
      tablePtr->tableFlags|=TBL_EDIT_ALLOWED;
      tablePtr->tableFlags&=~TBL_1_KEY_DESTROY;
      break;
    }
    /* reread the selection */
    TableGetSelection(tablePtr);
    TableConfigCursor(tablePtr);
    result=TCL_OK;
    break;

  case CMD_DRAWMODE:
    if(argc!=3)
    {
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], "\" drawmode mode", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    subcommand_retval=parse_command(tablePtr->interp, drawmode_commands, argv[2]);
    if(subcommand_retval==0)
    {
    /* not parsed break */
      result=TCL_ERROR;
    }
    else
    {
      /* just set the drawing mode and refresh the sreen */
      tablePtr->drawMode=subcommand_retval;
      TableInvalidate(tablePtr, 0, 0, Tk_Width(tablePtr->tkwin),
          Tk_Height(tablePtr->tkwin),0);
      result=TCL_OK;
    }
    break;
  
  case CMD_FLASH:
    if(argc<3)
    {
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], "\" flash arg ?arg?", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    subcommand_retval=parse_command(tablePtr->interp, flash_commands, argv[2]);
    /* what commmand */
    switch(subcommand_retval)
    {
    /* not parsed, break */
    case 0:
      result=TCL_ERROR;
      break;

    /* on or off */
    case FLASH_MODE:
      if(argc!=4 || Tcl_GetBoolean(tablePtr->interp, argv[3], &value)!=TCL_OK)
      {
        Tcl_AppendResult(  tablePtr->interp, "wrong # args : should be \"",
              argv[0], " flash mode on|off\"", (char *)NULL);
        result=TCL_ERROR;
        break;
      }
      /* set the mode and call the configuration routine */
      if(value)
      {
        tablePtr->tableFlags|=TBL_FLASH_ENABLED;
        TableFlashConfigure(tablePtr,1);
      }
      else
      {
        tablePtr->tableFlags&=~TBL_FLASH_ENABLED;
        TableFlashConfigure(tablePtr,0);
      }
      result=TCL_OK;
      break;

    case FLASH_TAG:
      {
        Tcl_HashEntry *entryPtr;
    
        if(argc!=4)
        {
          Tcl_AppendResult(  tablePtr->interp, "wrong # args : should be \"",
                argv[0], " flash tag <tagname>\"", (char *)NULL);
          result=TCL_ERROR;
          break;
        }

        /* is the tag name valid */
        if((entryPtr=Tcl_FindHashEntry(tablePtr->tagTable, argv[3]))==NULL)
        {
          Tcl_AppendResult(  tablePtr->interp, "invalid tag name \"",
                argv[3], "\"", (char *)NULL);
          result=TCL_ERROR;
          break;
        }

        /* put the tag pointer into the widget record */
        tablePtr->flashTag=(tagStruct *)Tcl_GetHashValue(entryPtr);
        TableFlashConfigure(tablePtr, 1);
        result=TCL_OK;
        break;
      }
    
    case FLASH_TIMEOUT:
      if(argc!=4 || Tcl_GetInt(tablePtr->interp, argv[3], &value)!=TCL_OK)
      {
        Tcl_AppendResult(  tablePtr->interp, "wrong # args : should be \"",
              argv[0], " flash timeout arg\"", (char *)NULL);
        result=TCL_ERROR;
        break;
      }
      tablePtr->flashTime=max(1,value);
      result=TCL_OK;
      break;
    } /* CMD_FLASH */
    break;

  case CMD_BATCH:
    if(argc!=3 || Tcl_GetBoolean(tablePtr->interp, argv[2], &value)!=TCL_OK)
    {
      Tcl_AppendResult(  tablePtr->interp, "wrong # args : should be \"",
            argv[0], " batch on|off\"", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    if(value)
      tablePtr->tableFlags|=TBL_BATCH_ON;
    else
      tablePtr->tableFlags&=~TBL_BATCH_ON;
    result=TCL_OK;
    break;

  case CMD_ROW_STRETCH:
  case CMD_COL_STRETCH:
    if(argc!=3)
    {
      Tcl_AppendResult( tablePtr->interp, "wrong # args should be \"",
            argv[0], (retval==CMD_ROW_STRETCH)?"row":"col","stretch mode", (char *)NULL);
      result=TCL_ERROR;
      break;
    }
    subcommand_retval=parse_command(tablePtr->interp, stretch_commands, argv[2]);
    if(subcommand_retval==0)
    {
    /* not parsed break */
      result=TCL_ERROR;
    }
    else
    {
      if(retval==CMD_ROW_STRETCH)
        tablePtr->rowStretchMode=subcommand_retval;
      else
        tablePtr->colStretchMode=subcommand_retval;

      TableAdjustParams(tablePtr);
      TableInvalidate(tablePtr, 0, 0, Tk_Width(tablePtr->tkwin),
          Tk_Height(tablePtr->tkwin),0);
      result=TCL_OK;
    }
    break;
  
    
  }
  Tk_Release(clientdata);
  return result;
}


static int TableTagCommand(Table *tablePtr, int argc, char *argv[]) {
  int result, retval, i, newEntry, maxval, value;
  int row, col;
  int x, y, width, height;
  tagStruct *tagPtr;
  Tcl_HashEntry *entryPtr, *scanPtr, *newEntryPtr, *oldEntryPtr;
  Tcl_HashTable *hashTblPtr;
  Tcl_HashSearch search;
  char buf[150], *keybuf;

  /* parse the next argument */
  retval=parse_command(tablePtr->interp, tag_commands, argv[2]);
  switch(retval)
  {
  /* failed to parse the argument, error */
  case 0:
    return TCL_ERROR;

    /* just print out the tag names */
  case TAG_NAMES:
    entryPtr=Tcl_FirstHashEntry(tablePtr->tagTable, &search);
    while(entryPtr!=NULL)
    {
      Tcl_AppendResult(  tablePtr->interp, 
            Tcl_GetHashKey(tablePtr->tagTable, entryPtr)," ",
            (char *)NULL);
      entryPtr=Tcl_NextHashEntry(&search);
    }
    return TCL_OK;

  /* delete a tag */
  case TAG_DELETE:
    /* any arguments */
    if(argc<4)
    {
      Tcl_AppendResult(  tablePtr->interp, "wrong # args: should be \"",
            argv[0], " tag delete tagName ?tagName ...?", (char *)NULL);
      return TCL_ERROR;
    }

    /* run through the remaining arguments */
    for(i=3;i<argc;i++)
    {
      result=TCL_OK;
      /* cannot delete the Title tag */
      if(strcmp(argv[i], "Title")==0 || strcmp(argv[i], "sel")==0)
      {
        Tcl_AppendResult(tablePtr->interp, "cannot delete ",argv[i],
             " tag", (char *)NULL);
        return TCL_ERROR;
      }
      if((entryPtr=Tcl_FindHashEntry(tablePtr->tagTable, argv[i]))!=NULL)
      {
        /* get the tag pointer */
        tagPtr=(tagStruct *)Tcl_GetHashValue(entryPtr);

        /* delete all references to this tag in rows */
        scanPtr=Tcl_FirstHashEntry(tablePtr->rowStyles, &search);
        for(;scanPtr!=NULL;scanPtr=Tcl_NextHashEntry(&search))
          if(Tcl_GetHashValue(scanPtr)==tagPtr)
            Tcl_DeleteHashEntry(scanPtr);

        /* delete all references to this tag in cols */
        scanPtr=Tcl_FirstHashEntry(tablePtr->colStyles, &search);
        for(;scanPtr!=NULL;scanPtr=Tcl_NextHashEntry(&search))
          if(Tcl_GetHashValue(scanPtr)==tagPtr)
            Tcl_DeleteHashEntry(scanPtr);

        /* delete all references to this tag in cells */
        scanPtr=Tcl_FirstHashEntry(tablePtr->cellStyles, &search);
        for(;scanPtr!=NULL;scanPtr=Tcl_NextHashEntry(&search))
          if(Tcl_GetHashValue(scanPtr)==tagPtr)
            Tcl_DeleteHashEntry(scanPtr);

        /* delete the reference in the flash tag */
        if(tablePtr->flashTag==tagPtr)
        {
          tablePtr->flashTag=NULL;
          TableFlashConfigure(tablePtr, 0);
        }

        /* release the structure */
        TableCleanupTag(tablePtr, (tagStruct *)Tcl_GetHashValue(entryPtr));
        ckfree((tagStruct *)Tcl_GetHashValue(entryPtr));

        /* and free the hash table entry */
        Tcl_DeleteHashEntry(entryPtr);
      }
    }
    /* since we deleted a tag, redraw the screen */
    TableInvalidate(tablePtr, 0, 0, Tk_Width(tablePtr->tkwin), Tk_Height(tablePtr->tkwin),0);
    return result;
    
  case TAG_CONFIGURE:
    if(argc<4)
    {
      Tcl_AppendResult(  tablePtr->interp, "wrong # args: should be \"",
            argv[0], " tag configure tagName ?arg arg  ...?", 
            (char *)NULL);
      return TCL_ERROR;
    }
      
    /* first see if this is a reconfiguration */
    entryPtr=Tcl_FindHashEntry(tablePtr->tagTable, argv[3]);
    if(entryPtr==NULL)
    {
      entryPtr=Tcl_CreateHashEntry(tablePtr->tagTable, argv[3], &newEntry);

      /* create the structure */
      tagPtr=TableNewTag(tablePtr);
        
      /* insert it into the table */
      Tcl_SetHashValue(entryPtr, (ClientData)tagPtr);

      /* configure the tag structure */
      result=Tk_ConfigureWidget(  tablePtr->interp, tablePtr->tkwin, tagConfig, 
                argc-4, argv+4, (char *)tagPtr, 0);
      if(result==TCL_ERROR)
        return TCL_ERROR;

      /* and create the graphics contexts etc. */
      TableTagConfig( tablePtr, tagPtr);    
    }
    /*
    ** pointer wasn't null, do a reconfig if we 
    ** have enough arguments
    */
    else 
    {
      /* get the tag pointer from the table*/
      tagPtr=(tagStruct *)Tcl_GetHashValue(entryPtr);
        
      /* 5 args means that there are values to replace */
      if(argc>5)
      {
        /* and do a reconfigure */
        result=Tk_ConfigureWidget(  tablePtr->interp, tablePtr->tkwin, tagConfig,
                argc-4, argv+4, (char*)tagPtr, TK_CONFIG_ARGV_ONLY);
        if(result==TCL_ERROR)
          return TCL_ERROR;
        /* and re-create the graphics contexts etc. */
        TableTagConfig( tablePtr, tagPtr);    
      }
    }
    /* 
    ** If there were less than 6 args, we need
    ** to do a printout of the config, even for
    ** new tags
    */
    if(argc<6)
    {
      result=Tk_ConfigureInfo(tablePtr->interp, tablePtr->tkwin, 
            tagConfig, (char *)tagPtr, 
            (argc==5) ? argv[4] : 0, 0);
    }
    /* 
    ** Otherwise we reconfigured so invalidate the table 
    ** for a redraw
    */
    else
      TableInvalidate(tablePtr, 0, 0, Tk_Width(tablePtr->tkwin), Tk_Height(tablePtr->tkwin),0);
    return result;
  

   /* tag a row or a column */
  case TAG_ROWTAG:
  case TAG_COLTAG:
    if (argc<4)
    {
      Tcl_AppendResult(tablePtr->interp, "wrong # args: should be \"", argv[0], 
          (retval==TAG_ROWTAG) ? " rowtag" : " coltag",
           " tag ?arg arg ..?\"", (char *)NULL);
      return TCL_ERROR;
    }

    /* if the tag is null, we are deleting */
    if(!(*argv[3]))
      tagPtr=NULL;
    
    else  /* check to see if the tag actually exists */
    {
      if((entryPtr=Tcl_FindHashEntry(tablePtr->tagTable, argv[3]))==NULL)
      {
        Tcl_AppendResult(tablePtr->interp, "invalid tag name \"", argv[3], "\"", (char *)NULL);
        return TCL_ERROR;  
      }
      /* get the pointer to the tag structure */
      tagPtr=(tagStruct *)Tcl_GetHashValue(entryPtr);
    }
  
    /* and choose the correct hash table */
    hashTblPtr=(retval==TAG_ROWTAG) ? tablePtr->rowStyles : tablePtr->colStyles;

    /* No more args -> display only */
    if(argc==4)
    {
      for(  scanPtr=Tcl_FirstHashEntry(hashTblPtr, &search);
        scanPtr!=NULL ; scanPtr=Tcl_NextHashEntry(&search))
        /* is this the tag pointer on this row */
        if((tagStruct *)Tcl_GetHashValue(scanPtr)==tagPtr)
        {
          sprintf(buf, "%d ", (int)Tcl_GetHashKey(hashTblPtr, scanPtr));
          Tcl_AppendResult(tablePtr->interp, buf, (char *)NULL);
        }
      return TCL_OK;
    }
    /* Now loop through the arguments and fill in the hash table */
    for(i=4;i<argc;i++)
    {
      /* can I parse this argument */
      if(  sscanf(argv[i], "%d", &value)!=1)
      {
        Tcl_AppendResult(tablePtr->interp, "bad value \"", argv[i], "\"", (char *)NULL);
        return TCL_ERROR;
      }

      /* deleting or adding */
      if(tagPtr==NULL)
      {
        oldEntryPtr=Tcl_FindHashEntry( hashTblPtr, (char *)value);
        if(oldEntryPtr!=NULL)
          Tcl_DeleteHashEntry(oldEntryPtr);
      }
      else
      {
        /* add a key to the hash table */
        newEntryPtr=Tcl_CreateHashEntry( hashTblPtr, (char *)value, &newEntry);
      
        /* and set it to point to the Tag structure */
        Tcl_SetHashValue(newEntryPtr, (ClientData)tagPtr);
      }
      /* and invalidate the row or column affected */
      if(retval==TAG_ROWTAG)
      {
        /* get the position of the leftmost cell in the row */
        TableCellCoords(tablePtr, value, 0, &x, &y, &width, &height);
        
        /* Invalidate the row */
        TableInvalidate(tablePtr, 0, y, Tk_Width(tablePtr->tkwin), height,0);
      }
      else
      {
        /* get the position of the topmost cell on the column */
        TableCellCoords(tablePtr, 0, value, &x, &y, &width, &height);
        
        /* Invalidate the column */
        TableInvalidate(tablePtr, x, 0, width, Tk_Height(tablePtr->tkwin),0);
      }
    }
    return TCL_OK;

   /* tag a cell */
  case TAG_CELLTAG:
    if (argc<4)
    {
      Tcl_AppendResult(tablePtr->interp, "wrong # args: should be \"", argv[0], 
          " celltag", " tag ?arg arg ..?\"", (char *)NULL);
      return TCL_ERROR;
    }

    /* are we deleting */
    if(!(*argv[3]))
      tagPtr=NULL;
    else
    {
      /* check to see if the tag actually exists */
      if((entryPtr=Tcl_FindHashEntry(tablePtr->tagTable, argv[3]))==NULL)
      {
        Tcl_AppendResult(tablePtr->interp, "invalid tag name \"", argv[3], "\"", (char *)NULL);
        return TCL_ERROR;  
      }
      /* get the pointer to the tag structure */
      tagPtr=(tagStruct *)Tcl_GetHashValue(entryPtr);
    }
  
    /* No more args -> display only */
    if(argc==4)
    {
      for(  scanPtr=Tcl_FirstHashEntry(tablePtr->cellStyles, &search);
        scanPtr!=NULL ; scanPtr=Tcl_NextHashEntry(&search))
        /* is this the tag pointer for this cell */
        if((tagStruct *)Tcl_GetHashValue(scanPtr)==tagPtr)
        {  
          keybuf=Tcl_GetHashKey(tablePtr->cellStyles, scanPtr);
    
          /* Split the value into its two components */
          TableParseArrayIndex(tablePtr, &row, &col, keybuf);
          sprintf(buf, "{%d %d} ", row, col);
          Tcl_AppendResult(tablePtr->interp, buf, (char *)NULL);
        }
      return TCL_OK;
    }
    /* Now loop through the arguments and fill in the hash table */
    for(i=4;i<argc;i++)
    {
      /* can I parse this argument */
      if(  sscanf(argv[i], "%d %d", &row, &col)!=2) 
      {
        Tcl_AppendResult(tablePtr->interp, "bad value \"", argv[i], "\"", (char *)NULL);
        return TCL_ERROR;
      }
      /* get the hash key ready */
      TableMakeArrayIndex(tablePtr, row, col, buf);
      
      /* is this a deletion */
      if(tagPtr==NULL)
      {
        oldEntryPtr=Tcl_FindHashEntry(tablePtr->cellStyles, buf);
        if(oldEntryPtr!=NULL)
          Tcl_DeleteHashEntry(oldEntryPtr);
      }
      else
      {
        /* add a key to the hash table */
        newEntryPtr=Tcl_CreateHashEntry( tablePtr->cellStyles, buf, &newEntry);
      
        /* and set it to point to the Tag structure */
        Tcl_SetHashValue(newEntryPtr, (ClientData)tagPtr);
      }
      /* now invalidate the area */
      TableCellCoords(tablePtr, row, col, &x, &y, &width, &height);
      TableInvalidate(tablePtr, x, y, width, height,0);
    }
    return TCL_OK;
  }
}



/* 
** Parses a command string passed in in arg comparing it with all the command strings
** in the command array. If it finds a string which is a unique identifier of one of the
** commands, returns the index . If none of the commands match, or the abbreviation is not 
** unique, then it sets up the message accordingly and returns 0
*/

static int parse_command(  Tcl_Interp *interp,  
        command_struct *commands, char *arg)
{
  int len=strlen(arg);
  command_struct *matched=(command_struct *)0;
  int err=0;
  command_struct *next=commands;

  while(*(next->name))
  {
    if(strncmp(next->name, arg, len)==0)
    {
      /* 
      ** have we already matched this one 
      ** if so make up an error message 
      */
      if(matched)
      {
        if(!err)
        {
          Tcl_AppendResult(  interp, "ambiguous option \"", arg,
                "\" could be ", matched->name, (char *)0);
          matched=next;
          err=1;
        }
        Tcl_AppendResult(interp, ", ", next->name, (char *)0);
      }
      else
        matched=next;
    }
    next++;
  }

  /* did we get multiple possibilities */
  if(err)
    return 0;

  /* did we match any at all */
  if(matched)
  {
    return matched->value;
  }
  else
  {
    Tcl_AppendResult(  interp, "bad option \"", arg,
          "\" must be ", (char *)NULL);
    next=commands;
    while(1)
    {
      Tcl_AppendResult(interp, next->name, (char *)NULL);

      /* the end of them all ?*/
      if(!*((++next)->name))
        return 0;

      /* or the last one at least */
      if(*((next+1)->name))
        Tcl_AppendResult(interp, ", ", (char *)NULL);
      else
        Tcl_AppendResult(interp, " or ", (char *)NULL);
    }
  }
}



/* 
** Get the current selection into the
** buffer and mark it as unedited
** set the position to the end of the string
*/
static void TableGetSelection(Table *tablePtr)
{
  char *data;
  char buf[100];

  TableMakeArrayIndex(  tablePtr, 
        tablePtr->selRow+tablePtr->rowOffset, tablePtr->selCol+tablePtr->colOffset, 
        buf) ;

  if  (tablePtr->arrayVar==NULL || 
    (data=Tcl_GetVar2(tablePtr->interp, tablePtr->arrayVar, buf, TCL_GLOBAL_ONLY))==NULL)
  {
    data="";
  }
  
  /* is the buffer long enough */
  TableBufLengthen(tablePtr, strlen(data));
  
  strcpy(tablePtr->selectBuf, data);
  tablePtr->textCurPosn=strlen(data);
  tablePtr->tableFlags &=~TBL_TEXT_CHANGED;
}


/* Turn row/col into an index into the table */
static void TableMakeArrayIndex(Table *tablePtr, int row, int col, char *buf) {
  if(tablePtr->rowThenCol)
    sprintf(buf, "%d,%d", row, col);
  else
    sprintf(buf, "%d,%d", col, row);
}

/* 
**Turn array index back into row/col 
** return the number of args parsed, 
** should be two 
*/
static int TableParseArrayIndex(Table *tablePtr, int *row, int *col, char *index) {
  if(tablePtr->rowThenCol)
       return sscanf(index, "%d,%d", row, col);
  else
       return sscanf(index, "%d,%d", col, row);
}

/*
** Parse the argument as an index into the
** selected cell text.
** recognises 'end' or an integer or insert
** constrains it to the size of the buffer 
*/
static int TableParseStringPosn(Table *tablePtr, char *arg, int *posn) {
    int temp;

  /* is this end */
  if(strncmp(arg, "end", min(strlen(arg), 3))==0)
    *posn=strlen(tablePtr->selectBuf);
  else if (strncmp(arg, "insert", min(strlen(arg), 6))==0)
    *posn=tablePtr->textCurPosn;
  else
  {
    if(sscanf(arg, "%d", &temp)!=1)
    {
      Tcl_AppendResult(tablePtr->interp, "error \"", arg, 
          "\" is not a valid index", (char *)NULL);    
      return TCL_ERROR;
    }
    *posn= min( max(0,temp) , strlen(tablePtr->selectBuf));
  }
  return TCL_OK;
}


/*
** Makes the selection buffer long enough to hold
** a string of length len PLUS an extra character
** for the \0. Copies over the data if it lengthens
** the buffer and frees the old one
*/
static void
TableBufLengthen(Table *tablePtr, int len)
{
  char *newbuf;
  if(tablePtr->selectBufLen<len+1)
  {
    tablePtr->selectBufLen=(int)(len*1.25)+8;
    newbuf=(char*)ckalloc(tablePtr->selectBufLen);
    if(tablePtr->selectBuf)
    {
      /* copy over the data */
      strcpy(newbuf, tablePtr->selectBuf);

      /* free the old buffer */
      ckfree(tablePtr->selectBuf);      
    }
    tablePtr->selectBuf=newbuf;
  }
}

/* 
** Calculate the row and column starts.
** Adjusts the topleft corner variable to keep
** it within the screen range, out of the titles
** and keep the screen full 
** make sure the selected cell is in the visible area
** checks to see if the top left cell has changed at all
** and invalidates the table if it has
*/

static void TableAdjustParams(Table *tablePtr) {
  int minX, maxX, minY, maxY,i,j;
  int topLeftCol, topLeftRow, botRightCol, botRightRow;
  int colPixelOffset, rowPixelOffset;
  int topRow, leftCol, total;
  int x, y, width, height;
  int diff, colPad, unpresetCols, lastUnpresetCol, lastColPad, setColPixels;
  int rowPad, unpresetRows, lastUnpresetRow, lastRowPad, setRowPixels;
  Tcl_HashEntry *entryPtr;
  int bd;


  /* make sure the top row and col have  reasonable values */
  tablePtr->topRow=topRow= max(tablePtr->titleRows, min(tablePtr->topRow, tablePtr->rows-1));
  tablePtr->leftCol=leftCol=max(tablePtr->titleCols, min(tablePtr->leftCol, tablePtr->cols-1));

  /* make sure the selected cell has a reasonable value */
  tablePtr->selRow=max(tablePtr->titleRows, min(tablePtr->selRow, tablePtr->rows-1));
  tablePtr->selCol=max(tablePtr->titleCols, min(tablePtr->selCol, tablePtr->cols-1));

  /* Set up the arrays to hold the row pixels and row starts */
  if(tablePtr->rowPixels)
    ckfree(tablePtr->rowPixels);
  tablePtr->rowPixels=(int *)calloc(tablePtr->rows, sizeof(int));

  if(tablePtr->rowStarts)
    ckfree(tablePtr->rowStarts);
  tablePtr->rowStarts=(int *)calloc(tablePtr->rows+1, sizeof(int));

  /* Set up the arrays to hold the row pixels and row starts */
  if(tablePtr->colPixels)
    ckfree(tablePtr->colPixels);
  tablePtr->colPixels=(int *)calloc(tablePtr->cols, sizeof(int));

  if(tablePtr->colStarts)
    ckfree(tablePtr->colStarts);
  tablePtr->colStarts=(int *)calloc(tablePtr->cols+1, sizeof(int));

  /* get the borderwidth for several upcoming calculations */
  bd=tablePtr->borderWidth;

  /* 
  ** get all the preset columns and set their widths
  */
  setColPixels=0;
  unpresetCols=0;
  total=0;
  for(i=0;i<tablePtr->cols;i++)
  {
    if((entryPtr=Tcl_FindHashEntry(tablePtr->colWidths, (char*)i))==NULL)
    {
      (tablePtr->colPixels)[i]=-1;
      unpresetCols++;
      lastUnpresetCol=i;
    }
    else
    {
      (tablePtr->colPixels)[i]=(int)Tcl_GetHashValue(entryPtr)*(tablePtr->charWidth)+2*bd;
      setColPixels+=(tablePtr->colPixels)[i];
    }
  }

  /* 
  ** get all the preset rows and set their heights
  */
  setRowPixels=0;
  unpresetRows=0;
  for(i=0;i<tablePtr->rows;i++)
  {
    if((entryPtr=Tcl_FindHashEntry(tablePtr->rowHeights, (char*)i))==NULL)
    {
      (tablePtr->rowPixels)[i]=-1;
      unpresetRows++;
      lastUnpresetRow=i;
    }
    else
    {
      (tablePtr->rowPixels)[i]=(int)Tcl_GetHashValue(entryPtr)+2*bd;
      setRowPixels+=(tablePtr->rowPixels)[i];
    }
  }

  /* 
  ** work out how much to pad each column and how much to
  ** pad the last column, depending on the mode 
  */
  if(  /*!Tk_IsMapped(tablePtr->tkwin) || */
    (tablePtr->colStretchMode==STRETCH_MODE_NONE) ||
    (tablePtr->colStretchMode==STRETCH_MODE_UNSET && unpresetCols==0) ||
    (tablePtr->cols==0))
  {
    colPad=0; lastColPad=0; 
  }
  else if(tablePtr->colStretchMode==STRETCH_MODE_UNSET)
  {
    diff=  max(0,(Tk_Width(tablePtr->tkwin)-setColPixels)-
      unpresetCols*((tablePtr->charWidth)*(tablePtr->defColWidth)+2*bd));
    colPad=diff/unpresetCols;
    lastColPad=diff-colPad*(unpresetCols-1);
  }
  else /* STRETCH_MODE_ALL */
  {
    diff=  max(0,(Tk_Width(tablePtr->tkwin)-setColPixels)-
      unpresetCols*((tablePtr->charWidth)*(tablePtr->defColWidth)+2*bd));
    colPad=diff/(tablePtr->cols);
    lastColPad=diff-colPad*(tablePtr->cols-1);

    /* force it to be applied to the last column too */
    lastUnpresetCol=tablePtr->cols-1;
  }
  
  /* now do the padding and calculate the column starts */
  total=0;
  for(i=0;i<tablePtr->cols;i++)
  {
    if((tablePtr->colPixels)[i]==-1)
      (tablePtr->colPixels)[i]=  (tablePtr->charWidth)*(tablePtr->defColWidth)+
              2*bd+((i!=lastUnpresetCol)?colPad:lastColPad);
    else if(tablePtr->colStretchMode==STRETCH_MODE_ALL)
      (tablePtr->colPixels)[i]+=  (i!=lastUnpresetCol)?colPad:lastColPad;

    total=(((tablePtr->colStarts)[i]=total)+(tablePtr->colPixels)[i]);
  }
  (tablePtr->colStarts)[i]=tablePtr->maxWidth= total;


  /* 
  ** work out how much to pad each row and how much to
  ** pad the last row, depending on the mode 
  */
  if(  /*!Tk_IsMapped(tablePtr->tkwin) || */
       (tablePtr->rowStretchMode==STRETCH_MODE_NONE) ||
    (tablePtr->rowStretchMode==STRETCH_MODE_UNSET && unpresetRows==0) ||
    (tablePtr->rows==0))
  {
    rowPad=0; lastRowPad=0; 
  }
  else if(tablePtr->rowStretchMode==STRETCH_MODE_UNSET)
  {
    diff=  max(0, (Tk_Height(tablePtr->tkwin)-setRowPixels)-
      unpresetRows*(tablePtr->defRowHeight+2*bd));
    rowPad=diff/unpresetRows;
    lastRowPad=diff-rowPad*(unpresetRows-1);
  }
  else /* STRETCH_MODE_ALL */
  {
    diff=  max(0, (Tk_Height(tablePtr->tkwin)-setRowPixels)-
      unpresetRows*(tablePtr->defRowHeight+2*bd));
    rowPad=diff/(tablePtr->rows);
    lastRowPad=diff-rowPad*(tablePtr->rows-1);

    /* force it to be applied to the last column too */
    lastUnpresetRow=tablePtr->rows-1;
  }
  
  /* now do the padding and calculate the row starts */
  total=0;
  for(i=0;i<tablePtr->rows;i++)
  {
    if((tablePtr->rowPixels)[i]==-1)
      (tablePtr->rowPixels)[i]=  (tablePtr->defRowHeight)+
              2*bd+((i!=lastUnpresetRow)?rowPad:lastRowPad);
    else if(tablePtr->colStretchMode==STRETCH_MODE_ALL)
      (tablePtr->colPixels)[i]+=  (i!=lastUnpresetRow)?rowPad:lastRowPad;

    /* calculate the start of each row */
    total=(((tablePtr->rowStarts)[i]=total)+(tablePtr->rowPixels)[i]);
  }
  (tablePtr->rowStarts)[i]=tablePtr->maxHeight= total;

  /* If we dont have the info, dont bother to fix up the other parameters */
  if(Tk_WindowId(tablePtr->tkwin)==None)
  {
    tablePtr->oldTopRow=tablePtr->oldLeftCol=-1;
    return;
  }

  /* make sure the top row and col have  reasonable values */
  tablePtr->topRow=topRow= max(tablePtr->titleRows, min(tablePtr->topRow, tablePtr->rows-1));
  tablePtr->leftCol=leftCol=max(tablePtr->titleCols, min(tablePtr->leftCol, tablePtr->cols-1));

  /* make sure the selected cell has a reasonable value */
  tablePtr->selRow=max(tablePtr->titleRows, min(tablePtr->selRow, tablePtr->rows-1));
  tablePtr->selCol=max(tablePtr->titleCols, min(tablePtr->selCol, tablePtr->cols-1));




  /* 
  ** constrain the top row and column to make the selected
  ** cell lie on the visible portion of the table if selection
  ** is enabled
  */
  if(tablePtr->selectionOn)
  {
    /* where is the new selected cell */
    TableCellCoords(tablePtr, tablePtr->selRow, tablePtr->selCol, &x, &y, &width, &height);
    
    if(x<(tablePtr->colStarts)[tablePtr->titleCols])
    {
      leftCol= tablePtr->selCol;
    }
    else if((x+width)>Tk_Width(tablePtr->tkwin))
    {
      int col;
      for(col=tablePtr->titleCols; col<tablePtr->selCol; col++)
      {  
        if( (tablePtr->colStarts)[tablePtr->titleCols]+
        ((tablePtr->colStarts)[tablePtr->selCol+1]-(tablePtr->colStarts)[col])<=Tk_Width(tablePtr->tkwin))
          break; 
      }
      leftCol=col;
    }

    if(y<(tablePtr->rowStarts)[tablePtr->titleRows])
      topRow=tablePtr->selRow;
    else if((y+height)>Tk_Height(tablePtr->tkwin))
    {
      int row;
      for(row=tablePtr->titleRows; row<tablePtr->selRow; row++)
      {
        if( (tablePtr->rowStarts)[tablePtr->titleRows]+
        ((tablePtr->rowStarts)[tablePtr->selRow+1]-(tablePtr->rowStarts)[row])<=Tk_Height(tablePtr->tkwin))
          break; 
      }
      topRow=row;
    }
  }

  /* 
  ** If we use this value of topRow, will we fill the window 
  ** if not, decrease it until we will, or until it gets to titleRows 
  ** make sure we dont cut off the bottom row
  */
  for(; topRow>tablePtr->titleRows; topRow--)
    if(  (tablePtr->maxHeight -
      ((tablePtr->rowStarts)[topRow-1]-(tablePtr->rowStarts)[tablePtr->titleRows])) >
      Tk_Height(tablePtr->tkwin))
      break;

  /* 
  ** If we use this value of topCol, will we fill the window 
  ** if not, decrease it until we will, or until it gets to titleCols 
  ** make sure we dont cut off the left column
  */
  for(; leftCol>tablePtr->titleCols; leftCol--)
    if(  (tablePtr->maxWidth - 
      ((tablePtr->colStarts)[leftCol-1]-(tablePtr->colStarts)[tablePtr->titleCols])) >
      Tk_Width(tablePtr->tkwin))
      break;

  tablePtr->topRow=topRow;
  tablePtr->leftCol=leftCol;

  /* 
  ** Do we have scrollbars, if so, calculate and call the TCL functions
  ** In order to get the scrollbar to be completely
  ** full when the whole screen is shown and there
  ** are titles, we have to arrange for the scrollbar
  ** range to be 0 -> rows-titleRows etc. This leads
  ** to the position setting methods, toprow and leftcol,
  ** being relative to the titles, not absolute row and
  ** column numbers.
  */
  if(tablePtr->yScrollCmd!=NULL || tablePtr->xScrollCmd!=NULL)
  {
    /* 
    ** Now work out where the bottom right  
    ** is so we can update the scrollbars
    */
    TableWhatCell(  tablePtr, 
        Tk_Width(tablePtr->tkwin)-1, Tk_Height(tablePtr->tkwin)-1, 
        &botRightRow, &botRightCol);

    /* Do we have a Y-scrollbar */
    if(tablePtr->yScrollCmd!=NULL)
    {
      char buffer[250];
      sprintf(buffer, " %d %d %d %d", tablePtr->rows-tablePtr->titleRows, max(1,botRightRow-topRow+1), 
        topRow-tablePtr->titleRows, max(topRow-tablePtr->titleRows+1, botRightRow-tablePtr->titleRows));
      if(Tcl_VarEval(  tablePtr->interp, 
          tablePtr->yScrollCmd, buffer, (char *)NULL)!=TCL_OK)
      {
        Tcl_AddErrorInfo(tablePtr->interp, 
            "\n(error executing yscroll command in table widget)");
        Tk_BackgroundError(tablePtr->interp);
      }
    }

    /* Do we have a X-scrollbar */
    if(tablePtr->xScrollCmd!=NULL)
    {
      char buffer[250];
      sprintf(buffer, " %d %d %d %d", tablePtr->cols-tablePtr->titleCols, max(1,botRightCol-leftCol+1), 
        leftCol-tablePtr->titleCols, max(leftCol-tablePtr->titleCols+1, botRightCol-tablePtr->titleCols));
      if(Tcl_VarEval(  tablePtr->interp, 
          tablePtr->xScrollCmd, buffer, (char *)NULL)!=TCL_OK)
      {
        Tcl_AddErrorInfo(tablePtr->interp, 
            "\n(error executing xscroll command in table widget)");
        Tk_BackgroundError(tablePtr->interp);
      }
    }
  }

  /* 
  ** now check the new value of topRow and 
  ** top column against the originals, did they
  ** even change. If they did, invalidate the
  ** area, if not, leave it alone
  */

  /* did the selected cell move then */
  if(  (tablePtr->selectionOn) && 
     (tablePtr->oldSelRow!=tablePtr->selRow || tablePtr->oldSelCol!=tablePtr->selCol))
  {
    int x, y, width, height; 
    char buf[100];
    TableCellCoords(tablePtr, tablePtr->oldSelRow, tablePtr->oldSelCol,
        &x, &y, &width, &height);
    TableInvalidate(tablePtr, x, y, width, height,1);
    TableCellCoords(tablePtr, tablePtr->selRow, tablePtr->selCol,
        &x, &y, &width, &height);
    TableInvalidate(tablePtr, x, y, width, height,1);

    /* put the value back in the cell */
    if(tablePtr->arrayVar!=NULL && (tablePtr->tableFlags & TBL_EDIT_ALLOWED))
    {
      char *tempvar;

      /* 
      ** if the value has not changed, do nothing
      ** else we will set off the 'flash' timer
      */
      
      TableMakeArrayIndex(  tablePtr, 
            tablePtr->oldSelRow+tablePtr->rowOffset, tablePtr->oldSelCol+tablePtr->colOffset, 
            buf);
      tempvar=Tcl_GetVar2(tablePtr->interp, tablePtr->arrayVar, buf, TCL_GLOBAL_ONLY);
      /* 
      ** if they are different, replace the value 
      ** the first check is: if the old value was null but the new
      ** buffer is not empty, replace, otherwise leave the value
      ** uninitialized
      */
      if(  (!tempvar && *(tablePtr->selectBuf)!='\0') ||
        ( tempvar  && strcmp(tempvar, tablePtr->selectBuf)))
        Tcl_SetVar2(tablePtr->interp, tablePtr->arrayVar, buf, tablePtr->selectBuf, TCL_GLOBAL_ONLY);
    }
  


    /* get the selection */
    TableGetSelection(tablePtr);
  }

  /* how about the rest of the table */
   if(tablePtr->topRow!=tablePtr->oldTopRow || tablePtr->leftCol!=tablePtr->oldLeftCol)
  {
    TableInvalidate(tablePtr, 0, 0, Tk_Width(tablePtr->tkwin), Tk_Height(tablePtr->tkwin),0);
  }

  /* 
  ** set the old top row and column for the next time this  
  ** function is called
  */
  tablePtr->oldTopRow=tablePtr->topRow; tablePtr->oldLeftCol=tablePtr->leftCol;
  tablePtr->oldSelRow=tablePtr->selRow; tablePtr->oldSelCol=tablePtr->selCol;
}


/*
** mallocs space for a new tag structure 
** and clears the structure returns 
** the pointer to the new structure
*/
static tagStruct * TableNewTag( Table *tablePtr ) {
  tagStruct *tagPtr;

  tagPtr=(tagStruct *)ckalloc(sizeof(tagStruct));
  tagPtr->configBgBorder=NULL;
  tagPtr->configForeground=NULL;
  tagPtr->configRelief=0;
  tagPtr->configFontPtr=NULL;
  tagPtr->configAnchor=-1;
  tagPtr->copyGc=None;
  
  return tagPtr;
}



/*
** This nasty piece of code sets up the GCs,
** colours, fonts etc.
** required for the tagStruct structure 
** for the window in the tablePtr
*/
static void TableTagConfig(Table *tablePtr, tagStruct *tagPtr) {
  XGCValues  gcValues;
  GC    newGC;

  XColor    *foreground;
  Tk_3DBorder  bgBorder;

  /* 
  ** If we are configuring a tag, some of the
  ** fields may be NULL, replace them with
  ** the default stuff for the table
  ** In this way all tags that 'don't care'
  ** will adjust when the main table adjusts
  */
  foreground=    (tagPtr->configForeground==NULL) ? 
        tablePtr->defaultTag.configForeground : tagPtr->configForeground;
  tagPtr->bgBorder=  bgBorder=(tagPtr->configBgBorder==NULL) ? 
        tablePtr->defaultTag.configBgBorder : tagPtr->configBgBorder;
  tagPtr->relief=    (tagPtr->configRelief==0) ? 
        tablePtr->defaultTag.configRelief : tagPtr->configRelief;
  tagPtr->fontPtr=  (tagPtr->configFontPtr==NULL) ? 
        tablePtr->defaultTag.configFontPtr : tagPtr->configFontPtr;
  tagPtr->anchor=    (tagPtr->configAnchor==-1) ? 
        tablePtr->defaultTag.configAnchor : tagPtr->configAnchor;
  /* 
  ** set up the graphics context for
  ** copying the pixmap onto the window
  */
  gcValues.foreground    =foreground->pixel;
  gcValues.background    =Tk_3DBorderColor(bgBorder)->pixel;
  gcValues.font      =tagPtr->fontPtr->fid;

  /* No exposures - this is from a pixmap */
  gcValues.graphics_exposures  =False;    

  /* get the GC */
  newGC=Tk_GetGC(  tablePtr->tkwin, GCForeground|GCBackground|GCFont|GCGraphicsExposures, 
      &gcValues);
  if(tagPtr->copyGc != None)
       Tk_FreeGC(tablePtr->display, tagPtr->copyGc);
  tagPtr->copyGc=newGC;
}


static void TableCleanupTag(Table *tablePtr, tagStruct *tagPtr) {
  /* free the graphics contexts */
   if(tagPtr->copyGc!=None)
    Tk_FreeGC(tablePtr->display, tagPtr->copyGc);

  /* free the options in the widget */
  Tk_FreeOptions(tagConfig, (char *)tagPtr, tablePtr->display, 0);
}



static void TableDisplay(ClientData clientdata) {
  Table *tablePtr    =(Table *)clientdata;
  Tk_Window tkwin    =tablePtr->tkwin;
  Display *display  =tablePtr->display;
  Tcl_Interp *interp  =tablePtr->interp;
  Drawable window;
  int i, j;
  int rowFrom, rowTo, colFrom, colTo;
  int firstBodyRow, firstBodyCol, lastBodyRow, lastBodyCol;
  int invalidX, invalidY, invalidWidth, invalidHeight;
  int x, y, width, height;
  int direction, ascent, descent, bd;
  int originX, originY;
  int selectedCell, clipRectSet;
  int offsetX, offsetY;
  XCharStruct bbox, bbox2;
  XRectangle  clipRect;
  GC topGc, bottomGc;
  char *value, buf[150];
  tagStruct *tagPtr;
  tagStruct *titlePtr;
  tagStruct *selPtr;
  Tcl_HashEntry *entryPtr;
  XPoint rect[6]={{0,0},{0,0},{0,0},{0,0},{0,0},{0,0}};

  tablePtr->tableFlags &= ~TBL_REDRAW_PENDING;
  invalidX=tablePtr->invalidX;
  invalidY=tablePtr->invalidY;
  invalidWidth=tablePtr->invalidWidth;
  invalidHeight=tablePtr->invalidHeight;

  /* 
  ** if we are using the slow draing mode with a pixmap 
  ** create the pixmap and set up offsetX and Y 
  */
  if(tablePtr->drawMode==DRAW_MODE_SLOW)
  {
    window=XCreatePixmap(  display, Tk_WindowId(tkwin),
          invalidWidth, invalidHeight, 
          Tk_Depth(tkwin));
    offsetX=invalidX; offsetY=invalidY;
  }
  else
  {
    window=Tk_WindowId(tablePtr->tkwin);
    offsetX=0; offsetY=0;
  }
  
  /* find the tag style for titles */
  entryPtr=Tcl_FindHashEntry(tablePtr->tagTable, "Title");
  titlePtr=(tagStruct *)Tcl_GetHashValue(entryPtr);
      
  /* find the tag type for the selection */
  entryPtr=Tcl_FindHashEntry(tablePtr->tagTable, "sel");
  selPtr=(tagStruct *)Tcl_GetHashValue(entryPtr);
      

  /* find out the cells represented by the invalid region */
  TableWhatCell(tablePtr, invalidX, invalidY, &rowFrom, &colFrom);
  TableWhatCell(tablePtr, invalidX+invalidWidth-1, invalidY+invalidHeight-1, &rowTo, &colTo);


  /* 
  ** Cycle through the cells and display them 
  */  
  for( i=rowFrom; i<=rowTo; i++)
    for(j=colFrom; j<=colTo ; j++)
    {
      /* 
      ** are we in the 'dead zone' between the
      ** titles and the first displayed cell 
      */
      if(i<tablePtr->topRow && i>=tablePtr->titleRows)
        i=tablePtr->topRow;      

      if(j<tablePtr->leftCol && j>=tablePtr->titleCols)
        j=tablePtr->leftCol;  

      /* are we in both row and column titles */
      if(i<tablePtr->titleRows && j<tablePtr->titleCols)
        continue;

      /* put the cell ref into a buffer for the hash lookups */
      TableMakeArrayIndex(tablePtr, i+tablePtr->rowOffset, j+tablePtr->colOffset,buf);

      /* get the tag structure for the cell */
      /* If there is a selected cell, is this it */
      if(selectedCell=((tablePtr->selectionOn) && 
           i==tablePtr->selRow && j==tablePtr->selCol))
        tagPtr=selPtr;
      /* if flash mode is on, is this cell flashing */
      else if(  (tablePtr->tableFlags&TBL_FLASH_ENABLED)&&
          (entryPtr=Tcl_FindHashEntry(tablePtr->flashCells, buf))!=NULL)
        tagPtr=tablePtr->flashTag;
      /* Am I in the titles */
      else if(i<tablePtr->topRow || j<tablePtr->leftCol)
        tagPtr=titlePtr;
      /* check the individual cell reference */
      else if((entryPtr=Tcl_FindHashEntry(tablePtr->cellStyles, buf))!=NULL)
        tagPtr=Tcl_GetHashValue(entryPtr);
      /* then try the rows */
      else if((entryPtr=Tcl_FindHashEntry(tablePtr->rowStyles, (char *)i+tablePtr->rowOffset))!=NULL)
        tagPtr=Tcl_GetHashValue(entryPtr);
      /* then the columns */
      else if((entryPtr=Tcl_FindHashEntry(tablePtr->colStyles, (char *)j+tablePtr->colOffset))!=NULL)
        tagPtr=Tcl_GetHashValue(entryPtr);
      /* it must be the default type then ! */
      else
        tagPtr=&(tablePtr->defaultTag);

      /* get the coordinates for the cell */
      TableCellCoords(tablePtr, i, j, &x, &y, &width, &height);

      /* 
      ** first fill in a blank rectangle. This is left
      ** as a Tk call instead of a direct X call for Tk
      ** compatibilty. The TK_RELIEF_FLAT ensures that only
      ** XFillRectangle is called anyway so the speed is the same
      */
      Tk_Fill3DRectangle(  display, window, tagPtr->bgBorder,
            x-offsetX, y-offsetY, width, height,
            tablePtr->borderWidth, TK_RELIEF_FLAT);   

      /* Is there a value in the cell? If so, draw it  */
        
      if(  (tablePtr->arrayVar!=NULL) &&   /* is there a variable at all */
        /* if selectTaed and editing, always show something */
        ((selectedCell && (tablePtr->tableFlags & TBL_EDIT_ALLOWED)) || 
        /* or there really is a value there */
        (value=Tcl_GetVar2(interp, tablePtr->arrayVar, buf, TCL_GLOBAL_ONLY))!=NULL))

      {  
        /* if this is the selected cell and we can edit, use the buffer */
        if(selectedCell && (tablePtr->tableFlags & TBL_EDIT_ALLOWED))
          value=tablePtr->selectBuf;

        /* get the border width to adjust the calculations*/
        bd=tablePtr->borderWidth;

        /* get the dimensions of the string */
        XTextExtents(  tagPtr->fontPtr, value, strlen(value),
            &direction, &ascent, &descent, &bbox);

        /* 
        ** Set the origin coordinates of the string to draw
        ** using the anchor point.
        ** origin represents the (x,y) coordinate of the
        ** lower left corner of the text box, relative to
        ** the internal (ie inside the border) window
        */
        
        /* set the X origin first */
        switch (tagPtr->anchor)
        {
        case TK_ANCHOR_NE: case TK_ANCHOR_E: case TK_ANCHOR_SE:
          originX=width-(bbox.rbearing-bbox.lbearing)-2*bd;
          break;
        case TK_ANCHOR_N: case TK_ANCHOR_S: case TK_ANCHOR_CENTER:
          originX=(width-(bbox.rbearing-bbox.lbearing))/2-bd;
          break;
        default:
          originX=0;
        }

        /* then set the Y origin */
        switch (tagPtr->anchor)
        {
        case TK_ANCHOR_N: case TK_ANCHOR_NE: case TK_ANCHOR_NW:
          originY=ascent+descent;
          break;
        case TK_ANCHOR_W: case TK_ANCHOR_E: case TK_ANCHOR_CENTER:
          originY=height-(height-(ascent+descent))/2-bd;
          break;
        default:
          originY=height-2*bd;
        }

        /*
        ** if this is the selected cell and we are editing
        ** ensure that the cursor will be displayed
        */
        if(selectedCell && (tablePtr->tableFlags & TBL_EDIT_ALLOWED))
        {
          XTextExtents(  tagPtr->fontPtr, value, 
              min(strlen(value), tablePtr->textCurPosn),
              &direction, &ascent, &descent, &bbox2);
          originX=max(3-bbox2.width, min(originX, width-bbox2.width-2*bd-3));
        }

        /* 
        ** only if I am desperate will I use a clip rectangle
        ** as it means updating the GC in the server which slows
        ** everything down.
        */
        if(   clipRectSet=
            (originX+bbox.lbearing<-bd || originX+bbox.rbearing > width - bd ||
            (ascent+descent > height)) )
        {
          /* set the clipping rectangle */
          clipRect.x=x-offsetX; clipRect.y=y-offsetY;
          clipRect.width=width; clipRect.height=height;
          XSetClipRectangles(display, tagPtr->copyGc, 0, 0, &clipRect, 1, Unsorted); 
        }

        XDrawString(  display, window, tagPtr->copyGc, 
            x-offsetX+originX+bd, y-offsetY+originY-descent+bd,
            value, strlen(value));

        /* if this is the selected cell draw the cursor if it's on */
        if(selectedCell && (tablePtr->tableFlags&TBL_CURSOR_ON))
        {
          Tk_Fill3DRectangle(  display, window, tablePtr->cursorBg, 
                x-offsetX+originX+bd+bbox2.width, 
                y-offsetY+originY-descent-ascent+bd,
                2, ascent+descent, 0,TK_RELIEF_FLAT); 
        }
        /* reset the clip mask */
        if(clipRectSet)
          XSetClipMask(display, tagPtr->copyGc, None); 
      }
      /* Draw the 3d border on the pixmap correctly offset */
      switch(tablePtr->drawMode)
      {
      case DRAW_MODE_SLOW:
      case DRAW_MODE_TK_COMPAT:
        Tk_Draw3DRectangle(  display, window, tagPtr->bgBorder,
              x-offsetX, y-offsetY, width, height,
              tablePtr->borderWidth, tagPtr->relief);     
        break;
      case DRAW_MODE_FAST:
        /* 
        ** choose the GCs to get the best approximation 
        ** to the desired drawing style
        */
        switch(tagPtr->relief)
        {
        case TK_RELIEF_FLAT:
          topGc=bottomGc=((Border *)(tagPtr->bgBorder))->bgGC;
          break;
        case TK_RELIEF_RAISED:
        case TK_RELIEF_RIDGE:
          topGc=((Border *)(tagPtr->bgBorder))->lightGC;
          bottomGc=((Border *)(tagPtr->bgBorder))->darkGC;
          break;
        default: /* TK_RELIEF_SUNKEN TK_RELIEF_GROOVE */
          bottomGc=((Border *)(tagPtr->bgBorder))->lightGC;
          topGc=((Border *)(tagPtr->bgBorder))->darkGC;
          break;
        }
        /* draw a line with no width */
        rect[0].x=x; rect[0].y=y+height-1; rect[1].y=-height+1; rect[2].x=width-1;
        rect[3].x=x+width-1;rect[3].y=y;rect[4].y=height-1;rect[5].x=-width  +1;
        XDrawLines( display, window, bottomGc, rect+3, 3, CoordModePrevious);  
        XDrawLines( display, window, topGc, rect, 3, CoordModePrevious);  
      }
    }

  /* clear the top left corner */
  if(tablePtr->titleRows!=0 || tablePtr->titleCols!=0)
       Tk_Fill3DRectangle(  display, window, titlePtr->bgBorder, 
          0-offsetX, 0-offsetY, (tablePtr->colStarts)[tablePtr->titleCols], 
               (tablePtr->rowStarts)[tablePtr->titleRows],
               (tablePtr->drawMode==DRAW_MODE_FAST)?1:tablePtr->borderWidth, titlePtr->relief);  



  /* 
  ** if we have got to the end of the table, 
  ** clear the area after the last row/column 
  */
  TableCellCoords(tablePtr, tablePtr->rows, tablePtr->cols, &x, &y, &width, &height);
  if(x+width<invalidX+invalidWidth-1)
    if(tablePtr->drawMode==DRAW_MODE_SLOW)
      /* we are using a pixmap, so use TK to clear it */
      Tk_Fill3DRectangle(   display, window, tablePtr->defaultTag.bgBorder, 
            x+width-offsetX, 0, invalidX+invalidWidth-x-width, invalidHeight,
            0, TK_RELIEF_FLAT);
    else  
      XClearArea(display, window, x+width, invalidY, invalidX+invalidWidth-x-width, invalidHeight, False);

  if(y+height<invalidY+invalidHeight-1)
    if(tablePtr->drawMode==DRAW_MODE_SLOW)
      /* we are using a pixmap, so use TK to clear it */
      Tk_Fill3DRectangle(   display, window, tablePtr->defaultTag.bgBorder, 
            0, y+height-offsetY, invalidWidth, invalidY+invalidHeight-y-height,
            0, TK_RELIEF_FLAT);
    else
      XClearArea(display, window, invalidX, y+height, invalidWidth,  invalidY+invalidHeight-y-height, False);

  /* copy on and delete the pixmap if we are in slow mode */
  if(tablePtr->drawMode==DRAW_MODE_SLOW)
  {
    XCopyArea(  display, window, Tk_WindowId(tkwin), 
        tablePtr->defaultTag.copyGc, 0, 0, 
        invalidWidth, invalidHeight, 
        invalidX, invalidY);
    XFreePixmap(Tk_Display(tkwin), window);
  }
}


/* 
** Returns the x, y, width and height for a 
** given cell
*/

static void TableCellCoords(Table *tablePtr, int row, int col, int *x, int *y, int *width, int *height) {
  int i, j, w=0, h=0;

  row=min(tablePtr->rows-1, max(0, row));
  col=min(tablePtr->cols-1, max(0, col));

  *width=(tablePtr->colPixels)[col];
  *height=(tablePtr->rowPixels)[row];
  *x=  (tablePtr->colStarts)[col] -
    ((col<tablePtr->titleCols)? 0: 
    (tablePtr->colStarts)[tablePtr->leftCol]-(tablePtr->colStarts)[tablePtr->titleCols]);
  *y=  (tablePtr->rowStarts)[row] -
    ((row<tablePtr->titleRows)? 0:
    (tablePtr->rowStarts)[tablePtr->topRow]-(tablePtr->rowStarts)[tablePtr->titleRows]);
}



static void TableWhatCell(Table *tablePtr, int x, int y, int *row, int *col) {
int i;

x=max(0, x); 
y=max(0, y); 

/* 
** Adjust the x coord if not in the column titles 
** to change display coords into internal coords
*/
x+=  (x<(tablePtr->colStarts)[tablePtr->titleCols]) ? 0 : 
  (tablePtr->colStarts)[tablePtr->leftCol]-(tablePtr->colStarts)[tablePtr->titleCols];
y+=  (y<(tablePtr->rowStarts)[tablePtr->titleRows]) ? 0 :
  (tablePtr->rowStarts)[tablePtr->topRow]-(tablePtr->rowStarts)[tablePtr->titleRows];

x=min(x, tablePtr->maxWidth-1);
y=min(y, tablePtr->maxHeight-1);

for(i=1;x>=(tablePtr->colStarts)[i];i++); *col=i-1;
for(i=1;y>=(tablePtr->rowStarts)[i];i++); *row=i-1;
}


static void  TableEventProc(ClientData clientdata, XEvent *eventPtr) {
  Table *tablePtr=(Table *)clientdata;

  switch (eventPtr->type)
  {
  case Expose:

    TableInvalidate(tablePtr, 
        eventPtr->xexpose.x, eventPtr->xexpose.y,
        eventPtr->xexpose.width, eventPtr->xexpose.height,0);
    break;

  case DestroyNotify:
    /* remove the command from the interpreter */
    Tcl_DeleteCommand(tablePtr->interp, Tk_PathName(tablePtr->tkwin));

    tablePtr->tkwin=NULL;
  
    /* cancel any pending update or timer */
    if(tablePtr->tableFlags & TBL_REDRAW_PENDING)
      Tk_CancelIdleCall(TableDisplay, (ClientData)tablePtr);
    if(tablePtr->cursorTimer!=NULL)
      Tk_DeleteTimerHandler(tablePtr->cursorTimer);

    /* delete the variable trace */
    if(tablePtr->arrayVar!=NULL)
      Tcl_UntraceVar(  tablePtr->interp, tablePtr->arrayVar, 
           TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
          TableVarProc, (ClientData) tablePtr);

    /* free the widget in time */  
    Tk_EventuallyFree((ClientData) tablePtr, TableDestroy);
    break;
  
  case ConfigureNotify:
    TableAdjustParams(tablePtr);
    TableInvalidate(tablePtr, 0, 0, Tk_Width(tablePtr->tkwin), Tk_Height(tablePtr->tkwin),0);
    break;

  case FocusIn:
    {
      int x, y, width, height;
      tablePtr->tableFlags |= TBL_HAS_FOCUS;
      
      /* Turn the cursor on */
      TableConfigCursor(tablePtr);
      break;
    }

  case FocusOut:
    tablePtr->tableFlags &= ~TBL_HAS_FOCUS;

    /* cancel the timer */
    TableConfigCursor(tablePtr);
    break;
  }
  
}


static void TableInvalidate(Table *tablePtr, int x, int y, int width, int height, int force) {

    int old_x, old_y, old_width, old_height;

  /* If not even mapped, ignore */
  if(!Tk_IsMapped(tablePtr->tkwin))
    return;

  /* is this rectangle even on the screen */
  if((x>Tk_Width(tablePtr->tkwin))||(y>Tk_Height(tablePtr->tkwin)))
    return;


  /* 
  ** if no pending updates then 
  ** replace the rectangle, otherwise    
  ** find the bounding rectangle
  */
  if(!(tablePtr->tableFlags & TBL_REDRAW_PENDING)) 
  {
    tablePtr->invalidX=x;
    tablePtr->invalidY=y;
    tablePtr->invalidWidth=width;
    tablePtr->invalidHeight=height;
    if(force && !(tablePtr->tableFlags & TBL_BATCH_ON))
      TableDisplay((ClientData)tablePtr);
    else
    {
      tablePtr->tableFlags |= TBL_REDRAW_PENDING;  
      Tk_DoWhenIdle(TableDisplay, (ClientData)tablePtr);  
    }
  }
  else
  {
    tablePtr->invalidWidth    =max(tablePtr->invalidX+tablePtr->invalidWidth, x+width);
    tablePtr->invalidHeight    =max(tablePtr->invalidY+tablePtr->invalidHeight, y+height);
    tablePtr->invalidX    =min(tablePtr->invalidX, x);
    tablePtr->invalidY    =min(tablePtr->invalidY, y);
    tablePtr->invalidWidth    -=tablePtr->invalidX;
    tablePtr->invalidHeight    -=tablePtr->invalidY;
    /* are we forcing this update out */
    if(force && !(tablePtr->tableFlags & TBL_BATCH_ON))
    {
      Tk_CancelIdleCall(TableDisplay, (ClientData)tablePtr);
      TableDisplay((ClientData)tablePtr);
    }      
  }
}



static void TableDestroy(ClientData clientdata) {
  Table *tablePtr=(Table *)clientdata;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;

  /* delete the variable trace */
  if(tablePtr->arrayVar!=NULL)
    Tcl_UntraceVar(  tablePtr->interp, tablePtr->arrayVar, 
         TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
        TableVarProc, (ClientData) tablePtr);
  
  /* destroy the row height and column width hash tables */
  if(tablePtr->colWidths!=NULL)
  {
    Tcl_DeleteHashTable(tablePtr->colWidths);
    ckfree(tablePtr->colWidths);
  }
  if(tablePtr->rowHeights!=NULL)
  {
    Tcl_DeleteHashTable(tablePtr->rowHeights);
    ckfree(tablePtr->rowHeights);
  }

  /* free the arrays with row/column pixel sizes */
  if(tablePtr->colPixels)
    ckfree(tablePtr->colPixels);
  if(tablePtr->rowPixels)
    ckfree(tablePtr->rowPixels);
  if(tablePtr->colStarts)
    ckfree(tablePtr->colStarts);
  if(tablePtr->rowStarts)
    ckfree(tablePtr->rowStarts);

  /* free the selection buffer */
  if(tablePtr->selectBuf)
    ckfree(tablePtr->selectBuf);

  /* delete the row, column and cell style hash tables */
  if(tablePtr->rowStyles!=NULL)
  {
    Tcl_DeleteHashTable(tablePtr->rowStyles);
    ckfree(tablePtr->rowStyles);
  }
  if(tablePtr->colStyles!=NULL)
  {
    Tcl_DeleteHashTable(tablePtr->colStyles);
    ckfree(tablePtr->colStyles);
  }
  if(tablePtr->cellStyles!=NULL)
  {
    Tcl_DeleteHashTable(tablePtr->cellStyles);
    ckfree(tablePtr->cellStyles);
  }

  if(tablePtr->flashCells!=NULL)
  {
    Tcl_DeleteHashTable(tablePtr->flashCells);
    ckfree(tablePtr->flashCells);
  }

  /* Now free up all the tag information */
  if(tablePtr->tagTable!=NULL)
  {
    entryPtr=Tcl_FirstHashEntry(tablePtr->tagTable, &search);    
    for(;entryPtr!=NULL;entryPtr=Tcl_NextHashEntry(&search))
    {
      /* free up the Gcs etc */
      TableCleanupTag(tablePtr, (tagStruct *)Tcl_GetHashValue(entryPtr));
  
      /* free the memory */
      ckfree((tagStruct *)Tcl_GetHashValue(entryPtr));
    }
    /* And delete the actual hash table */
    Tcl_DeleteHashTable(tablePtr->tagTable);
    ckfree(tablePtr->tagTable);
  }

  /* free up the stuff in the default tag */
  TableCleanupTag(tablePtr, &(tablePtr->defaultTag));

  /* free the configuration options in the widget */
  Tk_FreeOptions(TableConfig, (char *)tablePtr, tablePtr->display, 0);

  /* and free the widget memory at last! */
  ckfree(tablePtr);
}


static char * TableVarProc(  ClientData clientdata, Tcl_Interp *interp, char *name, char *index, int flags) {
  Table *tablePtr=(Table *)clientdata;
  int x, y, width, height, i, j;
  char buf[100];

  /* 
  ** is this the whole variable being destroyed 
  ** or just one cell being deleted
  */
  if(flags & TCL_TRACE_UNSETS && index==NULL)
  {
    /* 
    **if this isn't the interpreter being destroyed 
    ** reinstate the trace
    */
    if((flags & TCL_TRACE_DESTROYED) && !( flags & TCL_INTERP_DESTROYED))
    {
      Tcl_SetVar2(interp, tablePtr->arrayVar,  "0,0", "", TCL_GLOBAL_ONLY);
      Tcl_UnsetVar2(interp, tablePtr->arrayVar, "0,0", TCL_GLOBAL_ONLY);
      Tcl_ResetResult(interp);
      
      /* clear the selection buffer */
      TableGetSelection(tablePtr);

      /* set a trace on the variable */
      Tcl_TraceVar(  interp, tablePtr->arrayVar, 
          TCL_TRACE_WRITES | TCL_TRACE_UNSETS | TCL_GLOBAL_ONLY,
          TableVarProc, (ClientData) tablePtr);
      
      /* and invalidate the table */
      TableInvalidate(tablePtr, 0, 0, Tk_Width(tablePtr->tkwin), Tk_Height(tablePtr->tkwin),0);
    }
    return (char *)NULL;
  }
  
  /* get the cell address and invalidate that region only */
  TableParseArrayIndex(tablePtr, &i, &j, index);

  i-=tablePtr->rowOffset;
  j-=tablePtr->colOffset;

  /* did the selected cell just update */
  if(i==tablePtr->selRow&&j==tablePtr->selCol)
    TableGetSelection(tablePtr);

  /* Flash the cell */
  TableAddFlash(tablePtr, i, j);

  TableCellCoords(tablePtr, i, j, &x, &y, &width, &height);
  TableInvalidate(tablePtr, x, y, width, height,1);
  return (char *)NULL;
}



/*
** Toggle the cursor status 
*/
static void TableCursorEvent(ClientData clientdata) {
  Table *tablePtr=(Table *)clientdata;
  int x, y, width, height;

  /* Toggle the cursor */
  tablePtr->tableFlags ^= TBL_CURSOR_ON;
  
  /* Disable any old timers */
  if(tablePtr->cursorTimer!=NULL)
    /* 
    ** we ARE allowed to delete a timer that has 
    ** gone off.
    */
    Tk_DeleteTimerHandler(tablePtr->cursorTimer);

  /* set up a new timer */
  tablePtr->cursorTimer=Tk_CreateTimerHandler(500, TableCursorEvent, clientdata);

  /* invalidate the cell */
  TableCellCoords(tablePtr, tablePtr->selRow, tablePtr->selCol, &x, &y, &width, &height);
  TableInvalidate(tablePtr, x, y, width, height,1);
}


/*
** Configures the timer depending on the 
** state of the table.
*/
static void TableConfigCursor(Table *tablePtr) {
  int mask, x, y, width, height;

  mask=TBL_HAS_FOCUS | TBL_EDIT_ALLOWED;
  
  /* to get a cursor, we have to have focus, selection and edits allowed */
  if(((tablePtr->tableFlags & mask)==mask) && tablePtr->selectionOn)
  {
    /* turn the cursor on */
    tablePtr->tableFlags |= TBL_CURSOR_ON;
  
    /* set up the first timer */
    tablePtr->cursorTimer=Tk_CreateTimerHandler(500, TableCursorEvent, (ClientData)tablePtr);
  }
  else 
  {  
    /* turn the cursor off */
    tablePtr->tableFlags &= ~TBL_CURSOR_ON;
  
    /* and disable the timer */
    if(tablePtr->cursorTimer!=NULL)
      /* we ARE allowed to delete a timer that has gone off */
      Tk_DeleteTimerHandler(tablePtr->cursorTimer);
  }

  /* Invalidate the selection window to show or hide the cursor*/
  TableCellCoords(tablePtr, tablePtr->selRow, tablePtr->selCol,
      &x, &y, &width, &height);
  TableInvalidate(tablePtr, x, y, width, height,0); 
}


/* 
** starts or stops the timer for the flashing cells 
** if stopping, clears out the flash structure
*/
static void TableFlashConfigure(Table *tablePtr, int startStop) {
  if(startStop && (tablePtr->tableFlags&TBL_FLASH_ENABLED) && tablePtr->flashTag!=NULL)
  {
    if(tablePtr->flashTimer!=NULL)
      Tk_DeleteTimerHandler(tablePtr->flashTimer);
      tablePtr->flashTimer=Tk_CreateTimerHandler(100, TableFlashEvent, (ClientData)tablePtr);
  }
  else
  {
    
    if(tablePtr->flashTimer!=NULL)
      Tk_DeleteTimerHandler(tablePtr->flashTimer);
    tablePtr->flashTimer=(Tk_TimerToken)NULL;

    /* clear out the flash hash table */
    if(tablePtr->flashCells!=NULL)
      Tcl_DeleteHashTable(tablePtr->flashCells);

    /* set up a fresh blank one for the next time */
    Tcl_InitHashTable(tablePtr->flashCells, TCL_STRING_KEYS);
  }
}


/* 
** called when the flash timer goes off
** decrements all the entries in the hash 
** table and invalidates any cells that 
** expire, deleteing them from the table
** if the table is now empty, stops the timer
** else reenables it 
*/
static void TableFlashEvent(ClientData clientdata) {
  Table *tablePtr=(Table *)clientdata;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  int entries;
  int count;

  entries=0;
  for(  entryPtr=Tcl_FirstHashEntry(tablePtr->flashCells, &search);
    entryPtr!=NULL;
    entryPtr=Tcl_NextHashEntry(&search))
  {
    count=(int)Tcl_GetHashValue(entryPtr);
    if(--count<=0)
    {
      int row, col, x, y, width, height;

      /* get the cell address and invalidate that region only */
      TableParseArrayIndex(tablePtr, &row, &col, Tcl_GetHashKey(tablePtr->flashCells, entryPtr));

      row-=tablePtr->rowOffset;
      col-=tablePtr->colOffset;

      /* delete the entry from the table */
      Tcl_DeleteHashEntry(entryPtr);

      TableCellCoords(tablePtr, row, col, &x, &y, &width, &height);
      TableInvalidate(tablePtr, x, y, width, height,1);

    }
    else
    {
      Tcl_SetHashValue(entryPtr, (ClientData)count);
      entries++;
    }
  }

  /* do I need to restart the timer */
  if(entries)
    tablePtr->flashTimer=Tk_CreateTimerHandler(100, TableFlashEvent, (ClientData)tablePtr);
  else
    tablePtr->flashTimer=(Tk_TimerToken)NULL;
}  


/*
** adds a flash to the list with the default
** timeout. Only adds it if flashing is
** enabled and there is a flash ta defined
*/
static void TableAddFlash(Table *tablePtr, int row, int col)
{
  char buf[100];
  int dummy, x, y, width, height;
  Tcl_HashEntry *entryPtr;

  if(!(tablePtr->tableFlags&TBL_FLASH_ENABLED && tablePtr->flashTag!=NULL))
    return;
  
  /* create the array index */
  TableMakeArrayIndex(  tablePtr, 
        row+tablePtr->rowOffset, col+tablePtr->colOffset, 
        buf);

  /* add the flash to the hash table */
  entryPtr=Tcl_CreateHashEntry(tablePtr->flashCells, buf, &dummy);
  Tcl_SetHashValue(entryPtr, tablePtr->flashTime);

  /* now set the timer if it's not already going and invalidate the area */
  if(tablePtr->flashTimer==NULL)
    tablePtr->flashTimer=Tk_CreateTimerHandler(100, TableFlashEvent, (ClientData)tablePtr);
  
  /*TableCellCoords(tablePtr, row, col,
      &x, &y, &width, &height);
  TableInvalidate(tablePtr, x, y, width, height,); */
}


int Table_Init(Tcl_Interp *interp) {

    static char initCmd[] =
    " if [file exists $tk_library/TableInit.tcl] {\n\
      source $tk_library/TableInit.tcl\n\
    } else { \n\
      set msg \"can't find $tk_library/TableInit.tcl; perhaps you \"\n\
                append msg \"need to\\ninstall TableInit.tcl in your TK_LIBRARY directory?\"\n\
                append msg \"environment variable?\"\n\
                error $msg\n\
          }";

  Tcl_CreateCommand(  interp, "table", TableCmd,
        (ClientData) Tk_MainWindow(interp),
        (Tcl_CmdDeleteProc *)NULL);

  return Tcl_Eval(interp, initCmd);
  
}


/* 
** The command for creating a new instance of a
** Table widget
*/

int TableCmd(  ClientData clientdata, 
      Tcl_Interp *interp, 
      int argc, char *argv[])
{
  Tk_Window main_window= (Tk_Window) clientdata;
  Table     *tablePtr;  
  Tk_Window   tkwin;

  if (argc<2) 
  {
    Tcl_AppendResult(interp, "wrong #args: should be \"",
        argv[0], "pathname ?options?\"", (char *)NULL);
    return TCL_ERROR;
  }

  tkwin=Tk_CreateWindowFromPath(interp, main_window, argv[1], (char *)NULL);
  if (tkwin==NULL)
    return TCL_ERROR;
  Tk_SetClass(tkwin, "Table");

  tablePtr=(Table *)ckalloc(sizeof(Table));
  tablePtr->tkwin=tkwin;
  tablePtr->display=Tk_Display(tkwin);
  tablePtr->interp=interp;
  tablePtr->rows=0;
  tablePtr->cols=0;
  tablePtr->defRowHeight=0;
  tablePtr->defColWidth=0;
  tablePtr->arrayVar=NULL;
  tablePtr->borderWidth=0;
  tablePtr->defaultTag.configBgBorder=NULL;
  tablePtr->defaultTag.configForeground=NULL;
  tablePtr->defaultTag.configRelief=TK_RELIEF_FLAT;
  tablePtr->defaultTag.configFontPtr=NULL;
  tablePtr->defaultTag.configAnchor=TK_ANCHOR_CENTER;
  tablePtr->defaultTag.bgBorder=NULL;
  tablePtr->defaultTag.relief=TK_RELIEF_FLAT;
  tablePtr->defaultTag.fontPtr=NULL;
  tablePtr->defaultTag.anchor=TK_ANCHOR_CENTER;
  tablePtr->defaultTag.copyGc=None;
  tablePtr->yScrollCmd=NULL;
  tablePtr->xScrollCmd=NULL;
  tablePtr->cursorBg=NULL;
  tablePtr->rowThenCol=1;
  tablePtr->selectionOn=1;
  tablePtr->topRow=0;
  tablePtr->leftCol=0;
  tablePtr->titleRows=0;
  tablePtr->titleCols=0;
  tablePtr->selRow=0;
  tablePtr->selCol=0;
  tablePtr->oldTopRow=-1;
  tablePtr->oldLeftCol=-1;
  tablePtr->oldSelRow=-1;
  tablePtr->oldSelCol=-1;
  tablePtr->textCurPosn=0;
  tablePtr->tableFlags= TBL_EDIT_ALLOWED;
  tablePtr->drawMode=DRAW_MODE_SLOW;
  tablePtr->colStretchMode=STRETCH_MODE_NONE;
  tablePtr->rowStretchMode=STRETCH_MODE_NONE;
  tablePtr->maxWidth=0;
  tablePtr->maxHeight=0;
  tablePtr->charWidth=0;
  tablePtr->colPixels=(int *)0;
  tablePtr->rowPixels=(int *)0;
  tablePtr->colStarts=(int *)0;
  tablePtr->rowStarts=(int *)0;
  tablePtr->colOffset=0;
  tablePtr->rowOffset=0;
  tablePtr->flashTime=8;
  tablePtr->flashTag=(tagStruct *)0;
  tablePtr->colWidths=NULL;
  tablePtr->rowHeights=NULL;
  tablePtr->defaultTag.copyGc=None;
  tablePtr->tagTable=NULL;
  tablePtr->rowStyles=NULL;
  tablePtr->colStyles=NULL;
  tablePtr->cellStyles=NULL;
  tablePtr->flashCells=NULL;
  tablePtr->cursorTimer=(Tk_TimerToken)NULL;
  tablePtr->flashTimer=(Tk_TimerToken)NULL;
  tablePtr->selectBuf=(char*)ckalloc(1);*(tablePtr->selectBuf)='\0';
  tablePtr->selectBufLen=1;

  Tk_CreateEventHandler(  tkwin,
        ExposureMask | StructureNotifyMask | FocusChangeMask, TableEventProc,
        (ClientData) tablePtr);

  Tcl_CreateCommand(  interp, Tk_PathName(tkwin),
        TableWidgetCmd, (ClientData) tablePtr,
        (Tcl_CmdDeleteProc *)NULL);

  if (TableConfigure(interp, tablePtr, argc-2, argv+2, 0) != TCL_OK)
  {
    Tk_DestroyWindow(tkwin);
    return TCL_ERROR;
  }

  interp->result=Tk_PathName(tkwin);
  return TCL_OK;   
}

